"""
Utility methods and tasks for working with BigQuery/Google Cloud Storage from a Prefect flow.
"""

import os
from urllib.parse import urlparse

import backoff
import google.api_core.exceptions
from google.cloud import bigquery
from prefect import task
from prefect.utilities.gcp import get_bigquery_client, get_storage_client


@task
def cleanup_gcs_files(gcp_credentials: dict, url: str, project: str):
    """
    Task to delete files from a GCS prefix.

    Arguments:
      gcp_credentials (dict): GCP credentials in a format required by prefect.utilities.gcp.get_storage_client.
      url (str): Pointer to a GCS prefix containing one or more objects to delete.
      project (str): Name of the project which contains the target objects.
    """
    gcs_client = get_storage_client(credentials=gcp_credentials, project=project)
    parsed_url = urlparse(url)
    bucket = gcs_client.get_bucket(parsed_url.netloc)
    prefix = parsed_url.path.lstrip("/")
    # The list function is needed because bucket.list_blobs returns an
    # HTTPIterator object, which does not implement __len__.
    # But bucket.delete_blobs expects an Iterable with __len__.
    blobs = list(bucket.list_blobs(prefix=prefix))
    bucket.delete_blobs(blobs)
    return blobs


@task
@backoff.on_exception(backoff.expo,
                      google.api_core.exceptions.NotFound,
                      max_time=60*60*2)
def extract_ga_table(project: str, gcp_credentials: dict, dataset: str, date: str, output_root: str):
    """
    Runs a BigQuery extraction job, extracting the google analytics' `ga_sessions` table for a
    given date to a location in GCS in gzipped compressed JSON format.
    """
    table_name = "ga_sessions_{}".format(date)
    dest_filename = "{}_*.json.gz".format(table_name)
    base_extraction_path = os.path.join(output_root, dataset, date)
    destination_uri = os.path.join(base_extraction_path, dest_filename)

    client = get_bigquery_client(credentials=gcp_credentials, project=project)

    dataset = client.dataset(dataset, project=project)
    table = dataset.table(table_name)
    job_config = bigquery.job.ExtractJobConfig()
    job_config.destination_format = bigquery.DestinationFormat.NEWLINE_DELIMITED_JSON
    job_config.compression = "GZIP"
    extract_job = client.extract_table(table, destination_uri, job_config=job_config)
    extract_job.result()

    return base_extraction_path



"""
Utility methods for dealing with PayPal reports.
"""
import csv
import datetime
import fnmatch
import json

import prefect
from paramiko import SFTPClient, Transport
from prefect import config, task
from prefect.engine import signals

from edx_prefectutils.s3 import get_s3_path_for_date, list_object_keys_from_s3


def check_paypal_report(sftp_connection, remote_filename, check_column_name):
    """
    PayPal includes a row count metadata row to validate the report.
    - The type of a row is the first column of the row.
    - The number of rows with type 'SB' (the rows we care about) should equal the
      value of the single row of type 'SF'.
    - If there are 3 rows of type 'SB' then the second column of the 'SF' row should be 3.
    """
    report_file = sftp_connection.open(remote_filename, 'r')
    # First 3 lines are un-needed header
    file_read = report_file.readlines()[3:]
    reader = csv.DictReader(file_read, delimiter=',')

    # "CH" here is the column name of the first column in the report, it's effectively nonsense but consistent.
    #   It stands for "column header".
    # "SB" is the row type of rows we care about. It stands for "section body".
    # "SF" is the row type for summaries. It stands for "section footer".
    sb_count, sf_count = 0, 0
    for row in reader:
        if row['CH'] == 'SB':
            sb_count += 1
        if row['CH'] == 'SF':
            sf_count = int(row[check_column_name])
            break
    if sb_count != sf_count:
        raise Exception('Paypal row counts do not match for {}! Rows found: {}, Rows expected: {}'.format(
            remote_filename,
            sb_count,
            sf_count
        ))


def format_paypal_report(sftp_connection, remote_filename, date):
    """
    Removes unnecessary header / metadata rows that will mess up Snowflake loading.
    """
    report_file = sftp_connection.open(remote_filename, 'r')
    # First 3 lines are un-needed header
    file_read = report_file.readlines()[3:]
    reader = csv.DictReader(file_read, delimiter=',')

    output = []
    for row in reader:
        # "CH" here is the column name of the first column in the report, it's effectively nonsense but consistent.
        #   It stands for "column header".
        # "SB" is the row type of rows we care about. It stands for "section body".
        if row['CH'] == 'SB':
            row['report_date'] = date
            output.append(row)

    return json.dumps(output)


def get_paypal_filename(date, prefix, connection, remote_path):
    """
    Get remote filename. Sample remote filename: DDR-20190822.01.008.CSV
    """
    logger = prefect.context.get("logger")
    logger.info(connection)
    date_string = date.strftime('%Y%m%d')
    pattern = (prefix, date_string, 'CSV')
    remote_filepattern = "*".join(pattern)
    for file in connection.listdir(remote_path):
        # Ignore any file containing `_TEST` in its name.
        if fnmatch.fnmatch(file, remote_filepattern) and '_TEST' not in file:
            return file
    return None


class RemoteFileNotFoundError(Exception):
    pass


# Retry every 10 minutes for 5 hours! This should hopefully handle situations when the report is abnormally late.
@task(
    max_retries=30,
    retry_delay=datetime.timedelta(minutes=10),
    # Skip this retry filter until we upgrade to prefect 1.2.x since it is a new feature.
    retry_on=RemoteFileNotFoundError,
)
def fetch_paypal_report(
        date: str,
        paypal_credentials: dict,
        paypal_report_prefix: str,
        paypal_report_check_column_name: str,
        s3_bucket: str,
        s3_path: str,
        overwrite: bool,
):
    logger = prefect.context.get("logger")
    logger.info("Pulling Paypal report for {}".format(date))

    if not overwrite:
        # If we're not overwriting and the file already exists, raise a skip
        date_path = get_s3_path_for_date(date)
        s3_key = s3_path + date_path

        logger.info("Checking for existence of: {}".format(s3_key))

        existing_file = list_object_keys_from_s3.run(s3_bucket, s3_key)

        if existing_file:
            raise signals.SKIP(
                'File {} already exists and we are not overwriting. Skipping.'.format(s3_key)
            )
        else:
            logger.info("File not found, continuing download for {}.".format(date))

    transport = Transport(config.paypal.host, config.paypal.port)
    transport.connect(
        username=paypal_credentials.get('username'),
        password=paypal_credentials.get('password')
    )
    sftp_connection = SFTPClient.from_transport(transport)

    query_date = datetime.datetime.strptime(date, "%Y-%m-%d")
    remote_filename = get_paypal_filename(query_date, paypal_report_prefix, sftp_connection, config.paypal.remote_path)

    try:
        if remote_filename:
            sftp_connection.chdir(config.paypal.remote_path)
            check_paypal_report(sftp_connection, remote_filename, paypal_report_check_column_name)
            formatted_report = format_paypal_report(sftp_connection, remote_filename, date)
            return date, formatted_report
        else:
            raise RemoteFileNotFoundError("Remote File Not found for date: {0}".format(date))
    finally:
        sftp_connection.close()



"""
Tests for sitemap tasks.
"""

import json

import requests
from mock import Mock, patch
from prefect import context

from edx_prefectutils.sitemap import fetch_sitemap, fetch_sitemap_urls

SCRAPED_AT = '2021-10-22T15:14:16.683985+00:00'


@patch.object(requests, 'get')
def test_fetch_sitemap_urls(mockget):
    mockresponse = Mock()
    mockget.return_value = mockresponse
    mockresponse.text = """<?xml version="1.0" encoding="UTF-8"?>
    <sitemapindex xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
        <sitemap>
            <loc>https://www.foo.com/sitemap-0.xml</loc>
        </sitemap>
        <sitemap>
            <loc>https://www.foo.com/sitemap-1.xml</loc>
        </sitemap>
    </sitemapindex>
    """
    expected_output = ['https://www.foo.com/sitemap-0.xml', 'https://www.foo.com/sitemap-1.xml']
    task = fetch_sitemap_urls
    assert task.run(sitemap_index_url='dummy_url') == expected_output


@patch.object(requests, 'get')
def test_fetch_sitemap(mockget):
    mockresponse = Mock()
    mockget.return_value = mockresponse
    mockresponse.text = """<?xml version="1.0" encoding="UTF-8"?>
    <urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9" xmlns:video="http://www.google.com/schemas/sitemap-video/1.1"> # noqa: E501
        <url>
            <loc>https://www.foo.come/terms-service</loc>
            <changefreq>daily</changefreq>
            <priority>0.7</priority>
        </url>
        <url>
            <loc>https://www.foo.come/policy</loc>
            <changefreq>daily</changefreq>
            <priority>0.7</priority>
        </url>
        <url>
            <loc>https://www.foo.come/policy/security</loc>
            <changefreq>daily</changefreq>
            <priority>0.7</priority>
        </url>
    </urlset>
    """
    expected_output = [
        {'scraped_at': SCRAPED_AT, 'url': 'https://www.foo.come/terms-service'},
        {'scraped_at': SCRAPED_AT, 'url': 'https://www.foo.come/policy'},
        {'scraped_at': SCRAPED_AT, 'url': 'https://www.foo.come/policy/security'},
    ]

    task = fetch_sitemap
    with context(date=SCRAPED_AT):
        sitemap_filename, sitemap_json = task.run(sitemap_url='https://www.foo.com/sitemap-0.xml')
    assert (sitemap_filename, json.loads(sitemap_json)) == ('sitemap-0', expected_output)



#!/usr/bin/env python

"""
Tests for BigQuery utils in the `edx_prefectutils` package.
"""

from prefect.core import Flow
from pytest_mock import mocker  # noqa: F401

from edx_prefectutils import bigquery


def test_cleanup_gcs_files(mocker):  # noqa: F811
    mocker.patch.object(bigquery, 'get_storage_client')
    with Flow("test") as f:
        bigquery.cleanup_gcs_files(
            gcp_credentials={},
            url='',
            project='test_project'
        )
    state = f.run()
    assert state.is_successful()


def test_extract_ga_table(mocker):  # noqa: F811
    mocker.patch.object(bigquery, 'get_bigquery_client')
    with Flow("test") as f:
        task = bigquery.extract_ga_table(
            project='test_project',
            gcp_credentials={},
            dataset='test_dataset',
            date='2020-01-01',
            output_root='test_output_root'
        )
    state = f.run()
    assert state.is_successful()
    assert state.result[task].result == "test_output_root/test_dataset/2020-01-01"



"""
Tests for paypal SFTP.
"""

import datetime

from mock import Mock

from edx_prefectutils.paypal import get_paypal_filename


def test_get_paypal_filename():
    connection_mock = Mock()
    mock_list = [
        'TRR-20201209.01.009_TEST.CSV', 'TRR-20201209.01.009.CSV', 'TRR-20210114.01.009.CSV', 'TRR-20210115.01.011.CSV'
    ]
    connection_mock.listdir = Mock(return_value=mock_list)

    filename = get_paypal_filename(datetime.date(2020, 12, 9), 'TRR', connection_mock, 'dummy')
    assert filename == 'TRR-20201209.01.009.CSV'

    filename = get_paypal_filename(datetime.date(2021, 1, 15), 'TRR', connection_mock, 'dummy')
    assert filename == 'TRR-20210115.01.011.CSV'



"""
Test for S3 related tasks
"""

from mock import patch
from mock.mock import MagicMock

from edx_prefectutils.s3 import delete_s3_directory


@patch("edx_prefectutils.s3.list_object_keys_from_s3.run")
@patch("edx_prefectutils.s3.get_boto_client")
def test_delete_s3_directory(boto_client_mock, list_object_keys_from_s3_mock):
    """
    Test the delete_s3_directory task
    """
    keys = ["some/prefix/data_0_0_0.csv", "some/prefix/data_0_0_1.csv", "some/prefix/data_0_0_2.csv"]
    list_object_keys_from_s3_mock.return_value = keys
    client_mock = MagicMock()
    boto_client_mock.return_value = client_mock
    task = delete_s3_directory
    task.run(bucket="bucket", prefix="some/prefix/")

    client_mock.delete_objects.assert_called_once_with(
        Bucket="bucket",
        Delete={
            "Objects": [{"Key": key} for key in keys]
        },
    )



"""
Tasks for pulling sitemap data.
"""

import json
import xml.etree.ElementTree as ET
from os.path import basename, join, splitext
from urllib.parse import urlparse

import prefect
import requests
from prefect import task
from prefect.tasks.aws import s3


@task
def fetch_sitemap_urls(sitemap_index_url: str) -> str:
    """
    Fetches a list of sitemap urls by parsing sitemap index file.
    """
    r = requests.get(sitemap_index_url)
    sitemap_index_xml = r.text
    tree = ET.fromstring(sitemap_index_xml)
    sitemap_urls = [sitemap_node.find('{*}loc').text.strip() for sitemap_node in tree.findall('{*}sitemap')]
    return sitemap_urls


@task
def fetch_sitemap(sitemap_url: str):
    """
    Fetches sitemap data from a given sitemap URL.
    """
    scraped_at = str(prefect.context.date)
    r = requests.get(sitemap_url)
    sitemap_xml = r.text
    tree = ET.fromstring(sitemap_xml)

    sitemap_json = [
        {'scraped_at': scraped_at, 'url': url_node.find('{*}loc').text.strip()}
        for url_node in tree.findall('{*}url')
    ]

    sitemap_filename, ext = splitext(basename(urlparse(sitemap_url).path))
    return sitemap_filename, json.dumps(sitemap_json)


@task
def write_sitemap_to_s3(sitemap_data: str, s3_bucket: str, s3_path: str):
    """
    Writes sitemap data in JSON format to S3.
    """
    filename, sitemap_json = sitemap_data
    date_path = f'{prefect.context.today}/{filename}.json'
    s3_key = join(s3_path, date_path)
    s3.S3Upload(bucket=s3_bucket).run(
        sitemap_json,
        key=s3_key,
    )

    return date_path



#!/usr/bin/env python

"""
Tests for BigQuery utils in the `edx_prefectutils` package.
"""

from prefect.core import Flow

from edx_prefectutils import common


def test_generate_dates():
    with Flow("test") as f:
        task = common.generate_dates(
            start_date='2020-01-01',
            end_date='2020-01-05'
        )
    state = f.run()
    assert state.is_successful()
    assert state.result[task].result == ['20200101', '20200102', '20200103', '20200104']


def test_generate_month_start_dates():
    with Flow("test") as f:
        task = common.generate_month_start_dates(
            start_date='2023-01-31',
            end_date='2023-05-05'
        )
    state = f.run()
    assert state.is_successful()
    assert state.result[task].result == ['2023-01-01', '2023-02-01', '2023-03-01', '2023-04-01', '2023-05-01']


def test_get_unzipped_cartesian_product():
    with Flow("test") as f:
        task = common.get_unzipped_cartesian_product(
            input_lists=[[1, 2, 3], ["a", "b", "c"]]
        )
    state = f.run()
    assert state.is_successful()
    assert state.result[task].result == [
      (1, 1, 1, 2, 2, 2, 3, 3, 3),
      ("a", "b", "c", "a", "b", "c", "a", "b", "c")
    ]



import backoff
import prefect
import requests
from prefect import task
from prefect.backend import get_key_value, set_key_value
from snowflake.connector import ProgrammingError

from .snowflake import SFCredentials, get_batched_rows_from_snowflake

# Attribute arrays can only hold 25 items, though we could ask for this limit to be bumped. As of the time of this
# writing, our longest leads list is 14 items long (and only 4 users have more than 10). I think we're OK for now.
# https://www.braze.com/docs/api/objects_filters/user_attributes_object/
ARRAY_LIMIT = 25

# https://www.braze.com/docs/api/endpoints/user_data/post_user_track/
ATTRIBUTES_LIMIT = 75


def _make_custom_user_attrs(row):
    """Dict of attributes that are owned by this script"""
    # Masters are separated by semicolons in the snowflake database
    masters_leads = row.masters_leads.split(';') if row.masters_leads else []
    if len(masters_leads) > ARRAY_LIMIT:
        raise ValueError(
            f'At most {ARRAY_LIMIT} masters leads at a time can be set '
            f'in Braze, but we have {len(masters_leads)} for hubspot user {row.hs_object_id}'
        )

    # Programs are separated by semicolons in the snowflake database
    braze_programs = row.braze_programs.split(';') if row.braze_programs else []
    if len(braze_programs) > ARRAY_LIMIT:
        raise ValueError(
            f'At most {ARRAY_LIMIT} program interests at a time can be set '
            f'in Braze, but we have {len(braze_programs)} for hubspot user {row.hs_object_id}'
        )

    # Yes these are inconsistent. But it's a matter of historical field names.
    return {
        'HubSpot Audience': row.audience,
        'HubSpot Contact': row.braze_hs_contact,
        'HubSpot Contact ID': row.hs_object_id,
        'HubSpot Create Date': row.createdate.isoformat(),
        'HubSpot - First Conversion': row.first_conversion_event_name,
        'HubSpot - Lead Source': row.lead_source_deal_value,
        'HubSpot - Lead Source Detail': row.lead_source,  # yes, the bare `lead_source` is actually the detail
        'HubSpot Level of Education': row.highest_level_of_education,
        'HubSpot Masters Interests': masters_leads,
        'HubSpot Program Interests': braze_programs,
        'HubSpot - utm_campaign': row.utm_campaign,
        'HubSpot - utm_medium': row.utm_medium,
        'HubSpot - utm_source': row.utm_source,
    }


def _make_standard_user_attrs(row):
    """Dict of attributes that are standard braze ones - that we likely only need to set for unregistered users"""
    return {
        'country': row.country_iso_code,
        'email': row.email,
        'first_name': row.firstname,
        'home_city': row.city,
        'last_name': row.lastname,
        'phone': row.phone,
    }


@task
@backoff.on_exception(backoff.expo,
                      ProgrammingError,
                      max_tries=3)
def sync_hubspot_leads_to_braze(
    sf_credentials: SFCredentials,
    sf_role: str,
    sf_database: str,
    sf_schema: str,
    braze_api_key: str,
    braze_api_server: str,
):
    """
    Copies a bunch of hubspot data fields over to matching braze users.

    Note that we must handle users that don't have a proper braze external_id, because they are not registered.
    """
    logger = prefect.context.get('logger')
    columns = [
        'audience',
        'braze_hs_contact',
        'braze_programs',
        'city',
        'country_iso_code',
        'createdate',
        'email',
        'first_conversion_event_name',
        'firstname',
        'highest_level_of_education',
        'hs_object_id',
        'lastname',
        'lastmodifieddate',
        'lead_source',
        'lead_source_deal_value',
        'lms_user_id',
        'masters_leads',
        'phone',
        'utm_campaign',
        'utm_medium',
        'utm_source',
    ]

    # There are two types of users:
    # - Hubspot-alias-only users that don't have an account on the LMS
    # - Users that have an lms_user_id (maybe never had hubspot, maybe did - 'identifying' aliases is done by the
    #   platform-plugin-braze plugin that calls Braze's 'users/identify' API as users register on the LMS)
    #
    # We only want to set the "basic" fields like name/phone/email for alias-only users. Registered users get those
    # fields from the LMS. But both categories get all the hubspot-specific fields.

    alias_only_where = 'lms_user_id IS NULL'
    registered_where = 'lms_user_id IS NOT NULL'

    # Also, to reduce churn and data usage, only sync items that have changed since our last successful run
    LAST_SUCCESS_KEY = 'hubspot-leads-last-success'
    try:
        last_success = get_key_value(key=LAST_SUCCESS_KEY)
        alias_only_where += f" AND lastmodifieddate > '{last_success}'"
        registered_where += f" AND lastmodifieddate > '{last_success}'"
        logger.info('Read lastmodifieddate of %s', last_success)
    except ValueError:
        pass

    def _get_batches(label, where=None):
        count = 0
        for rows in get_batched_rows_from_snowflake(sf_credentials, sf_database, sf_schema, 'hubspot_leads', sf_role,
                                                    columns, ATTRIBUTES_LIMIT, where=where):
            count += len(rows)
            yield rows
            if count % (ATTRIBUTES_LIMIT * 10) == 0:
                logger.info(label, count)
        if count > 0:
            logger.info(label, count)

    max_date = None

    for batch in _get_batches('Updated %d hubspot alias users in Braze', where=alias_only_where):
        update_alias_users(braze_api_server, braze_api_key, batch)
        max_batch_date = max(row.lastmodifieddate for row in batch)
        max_date = max(max_batch_date, max_date) if max_date else max_batch_date

    for batch in _get_batches('Updated %d registered users in Braze', where=registered_where):
        update_registered_users(braze_api_server, braze_api_key, batch)
        max_batch_date = max(row.lastmodifieddate for row in batch)
        max_date = max(max_batch_date, max_date) if max_date else max_batch_date

    if max_date:
        logger.info('Saving lastmodifieddate of %s', max_date.isoformat())
        set_key_value(key=LAST_SUCCESS_KEY, value=max_date.isoformat())


# Retry rate limit errors for 5 minutes
@backoff.on_predicate(
    backoff.expo,
    lambda resp: resp.status_code == 429,
    max_time=300
)
# Retry on server error 3 times
@backoff.on_predicate(
    backoff.expo,
    lambda resp: resp.status_code >= 500,
    max_tries=3
)
def update_registered_users(braze_api_server, braze_api_key, rows):
    """
    Sets custom braze attributes on a list of registered users (users with an external_id set).

    Args:
        braze_api_server (str)
        braze_api_key (str)
        rows (list): sequence of user info
    """
    if len(rows) > ATTRIBUTES_LIMIT:
        raise ValueError(
            f'At most {ATTRIBUTES_LIMIT} users at a time can be updated in Braze, but we attempted {len(rows)}'
        )

    return requests.post(
        url=f"{braze_api_server}/users/track",
        headers={'Authorization': f'Bearer {braze_api_key}'},
        json={
            'attributes': [
                # https://www.braze.com/docs/api/objects_filters/user_attributes_object/
                {
                    'external_id': row.lms_user_id,
                    **_make_custom_user_attrs(row),
                } for row in rows
            ],
        },
    )


# Retry rate limit errors for 5 minutes
@backoff.on_predicate(
    backoff.expo,
    lambda resp: resp.status_code == 429,
    max_time=300
)
# Retry on server error 3 times
@backoff.on_predicate(
    backoff.expo,
    lambda resp: resp.status_code >= 500,
    max_tries=3
)
def update_alias_users(braze_api_server, braze_api_key, rows):
    """
    Update attributes for the given hubspot-alias-only users in braze, which don't have an external_id set.

    Args:
        braze_api_server (str)
        braze_api_key (str)
        rows (list): sequence of user info
    """
    if len(rows) > ATTRIBUTES_LIMIT:
        raise ValueError(
            f'At most {ATTRIBUTES_LIMIT} users at a time can be updated in Braze, but we attempted {len(rows)}'
        )

    return requests.post(
        url=f'{braze_api_server}/users/track',
        headers={'Authorization': f'Bearer {braze_api_key}'},
        json={
            'attributes': [
                # https://www.braze.com/docs/api/objects_filters/user_attributes_object/
                {
                    '_update_existing_only': False,
                    'user_alias': {
                        'alias_label': 'hubspot',
                        'alias_name': row.email,
                    },
                    **_make_standard_user_attrs(row),
                    **_make_custom_user_attrs(row),
                } for row in rows
            ],
        },
    )



"""
Tasks for fetching secrets from Vault.  There is a separate class per Vault
"engine", and they should all extend VaultSecretBase which centralizes the
logic to construct a Vault client from a Service Account JWT token.

Example usage for VaultKVSecret:

1. Create a KV secret at <https://vault.analytics.edx.org/ui/vault/secrets/kv/list>.

2. If, for example, the new secret had a path of
"snowflake_pipeline_etl_loader", use the following code pattern to use the
secret in Prefect flows:

    from vault_secrets import VaultKVSecret

    @task
    def load_data_into_snowflake(sf_credentials):
        self.logger.info("logging into Snowflake with username: {}".format(credentials["user"]))
        connection = create_snowflake_connection(credentials, role)
        ...
        connection.close()

    with Flow("Load Data Into Snowflake") as flow:
        sf_credentials = VaultKVSecret(
            path="snowflake_pipeline_etl_loader",
            version=3,
        )
        load_data_into_snowflake(sf_credentials)
"""
import hvac
from prefect.tasks.secrets import SecretBase

# This is a standardized k8s path to always find the service account JWT token.
SERVICE_ACCOUNT_JWT_TOKEN_PATH = "/var/run/secrets/kubernetes.io/serviceaccount/token"

# Global configuration for accessing vault from all Prefect flows inside of the analytics k8s
# cluster.
VAULT_BASE_URL = "https://vault.analytics.edx.org"
VAULT_LOGIN_URL = VAULT_BASE_URL + "/v1/auth/kubernetes/login"
VAULT_ROLE = "prefect"

# Global configuration for accessing vault from Prefect flows _outside_ of K8s.
EXTERNAL_VAULT_BASE_URL = "https://vault.analytics.edx.org"


class VaultSecretBase(SecretBase):
    """
    A base Secret task that establishes the vault client which can be used by
    extending classes to fetch secrets from Vault.

    Extending classes should override self._get_secret(), and use
    self.vault_client (instance of hvac.Client) to make requests.

    Args:
        **kwargs (Any, optional): additional keyword arguments to pass to the Task constructor
    """

    @staticmethod
    def _get_k8s_vault_client() -> hvac.Client:
        """
        Convert the current container's service account JWT token into a Vault
        token and use that to construct a Vault client.
        """
        with open(SERVICE_ACCOUNT_JWT_TOKEN_PATH) as sa_token_file:
            service_account_token = sa_token_file.read()
        client = hvac.Client(url=VAULT_BASE_URL)
        client.auth_kubernetes(role=VAULT_ROLE, jwt=service_account_token)
        return client

    @staticmethod
    def _get_env_var_vault_client() -> hvac.Client:
        """
        For local development, if the user is logged into Vault we can use
        their existing environment vars.
        """
        client = hvac.Client(url=EXTERNAL_VAULT_BASE_URL)
        return client

    def run(self):
        # First try k8s auth, if we can't find the magic file then try local env var authentication.
        try:
            self.vault_client = self._get_k8s_vault_client()
        except FileNotFoundError:
            # If that fails try local token auth
            self.vault_client = self._get_env_var_vault_client()

        if not self.vault_client.is_authenticated():
            raise Exception("Vault Client error. We don't seem to be in K8s and no Vault token found. "
                            "Try 'vault login -address https://vault.analytics.edx.org -method oidc' "
                            "if you've downloaded Vault.")

        return self._get_secret()

    def _get_secret(self):
        """
        Override this in extending classes to fetch the secret using
        `self.vault_client`.
        """
        pass


class VaultKVSecret(VaultSecretBase):
    """
    A `Secret` prefect task for fetching KV secrets from Vault. Note that this
    only supports version 2 of the KV engine.

    Manage KV secrets at https://vault.analytics.edx.org/ui/vault/secrets/kv/list

    Args:
        path (str): The path of the KV secret, e.g. "snowflake_pipeline_etl_loader".
        version (int): The version number of the KV secret.
        **kwargs (Any, optional): Additional keyword arguments to pass to the Task constructor.
    """

    def __init__(self, path: str, version: int, **kwargs):
        self.kv_path = path
        self.kv_version = version
        super().__init__(**kwargs)

    def _get_secret(self):
        """
        Fetch the KV secret specified by path and version.

        Returns:
            dict containing the key/value pairs, where the values are secrets.
        """
        secret_version_response = self.vault_client.secrets.kv.v2.read_secret_version(
            mount_point="kv", path=self.kv_path, version=self.kv_version,
        )
        return secret_version_response["data"]["data"]



#!/usr/bin/env python

"""
Tests for Snowflake utils in the `edx_prefectutils` package.
"""

import json

import mock
import pytest
from prefect.core import Flow
from prefect.engine import signals
from prefect.utilities.debug import raise_on_exception
from pytest_mock import mocker  # noqa: F401
from snowflake.connector import ProgrammingError

from edx_prefectutils import snowflake


def test_qualified_table_name():
    assert 'test_db.test_schema.test_table' == snowflake.qualified_table_name(
        'test_db', 'test_schema', 'test_table'
    )


def test_qualified_stage_name():
    assert 'test_db.test_schema.test_table_stage' == snowflake.qualified_stage_name(
        'test_db', 'test_schema', 'test_table'
    )


@pytest.fixture
def mock_sf_connection(mocker):  # noqa: F811
    # Mock the Snowflake connection and cursor.
    mocker.patch.object(snowflake, 'create_snowflake_connection')
    mock_cursor = mocker.Mock()
    mock_connection = mocker.Mock()
    mock_connection.cursor.return_value = mock_cursor
    snowflake.create_snowflake_connection.return_value = mock_connection
    return mock_connection


def test_create_snowflake_connection(mocker):  # noqa: F811
    # Mock out the snowflake connection method w/o mocking out the helper method.
    mocker.patch.object(snowflake.snowflake.connector, 'connect')
    mock_cursor = mocker.Mock()
    mock_connection = mocker.Mock()
    mock_connection.cursor.return_value = mock_cursor
    snowflake.snowflake.connector.connect.return_value = mock_connection
    # Mock the key decryption.
    mocker.patch.object(snowflake.serialization, 'load_pem_private_key')
    mock_key = mocker.Mock()
    mock_key.private_bytes.return_value = 1234
    snowflake.serialization.load_pem_private_key.return_value = mock_key
    # Call the connection method.
    snowflake.create_snowflake_connection(
        credentials={
            "private_key": "this_is_an_encrypted_private_key",
            "private_key_passphrase": "passphrase_for_the_private_key",
            "user": "test_user",
            "account": "company-cloud-region"
        },
        role="test_role"
    )
    snowflake.snowflake.connector.connect.assert_called_with(
        account='company-cloud-region',
        autocommit=False,
        private_key=1234,
        user='test_user',
        warehouse=None,
        password=None,
    )
    mock_cursor.execute.assert_has_calls(
        (
            mocker.call("USE ROLE test_role"),
            mocker.call("ALTER SESSION SET TIMEZONE = 'UTC'"),
        )
    )


def test_load_json_objects_to_snowflake_no_existing_table(mock_sf_connection):
    # Mock the Snowflake connection, cursor, and fetchone method.
    mock_cursor = mock_sf_connection.cursor()
    mock_fetchone = mock.Mock(side_effect=ProgrammingError("does not exist"))
    mock_cursor.fetchone = mock_fetchone

    with Flow("test") as f:
        snowflake.load_ga_data_to_snowflake(
            sf_credentials={},
            sf_database="test_database",
            sf_schema="test_schema",
            sf_table="test_table",
            sf_role="test_role",
            sf_warehouse="test_warehouse",
            sf_storage_integration="test_storage_integration",
            bq_dataset="test_dataset",
            gcs_url="gs://test-location",
            date="2020-01-01",
        )
    state = f.run()
    assert state.is_successful()
    mock_cursor.execute.assert_has_calls(
        [
            mock.call("\n        SELECT 1 FROM test_database.test_schema.test_table\n        WHERE session:date='2020-01-01'\n            AND ga_view_id='test_dataset'\n        "), # noqa
            mock.call('\n        CREATE TABLE IF NOT EXISTS test_database.test_schema.test_table (\n            id number autoincrement start 1 increment 1,\n            load_time timestamp_ltz default current_timestamp(),\n            ga_view_id string,\n            session VARIANT\n        );\n        '), # noqa
            mock.call("\n        CREATE OR REPLACE STAGE test_database.test_schema.test_table_stage\n            URL = 'gcs://test-location'\n            STORAGE_INTEGRATION = test_storage_integration\n            FILE_FORMAT = (TYPE = JSON);\n        "), # noqa
            mock.call("\n        COPY INTO test_database.test_schema.test_table (ga_view_id, session)\n            FROM (\n                SELECT\n                    'test_dataset',\n                    t.$1\n                FROM @test_database.test_schema.test_table_stage t\n            )\n        PATTERN='.*'\n        FORCE=False\n        "), # noqa
        ]
    )


def test_load_json_objects_to_snowflake_error_on_table_exist_check(mock_sf_connection):
    # Mock the Snowflake connection, cursor, and fetchone method.
    mock_cursor = mock_sf_connection.cursor()
    mock_fetchone = mock.Mock(side_effect=ProgrammingError())
    mock_cursor.fetchone = mock_fetchone

    with Flow("test") as f:
        snowflake.load_ga_data_to_snowflake(
            sf_credentials={},
            sf_database="test_database",
            sf_schema="test_schema",
            sf_table="test_table",
            sf_role="test_role",
            sf_warehouse="test_warehouse",
            sf_storage_integration="test_storage_integration",
            bq_dataset="test_dataset",
            gcs_url="gs://test-location",
            date="2020-01-01",
        )
    with raise_on_exception():
        with pytest.raises(ProgrammingError):
            f.run()


def test_load_json_objects_to_snowflake_overwrite(mock_sf_connection):
    # Mock the Snowflake connection, cursor, and fetchone method.
    mock_cursor = mock_sf_connection.cursor()
    mock_fetchone = mock.Mock(return_value=None)
    mock_cursor.fetchone = mock_fetchone

    with Flow("test") as f:
        snowflake.load_ga_data_to_snowflake(
            sf_credentials={},
            sf_database="test_database",
            sf_schema="test_schema",
            sf_table="test_table",
            sf_role="test_role",
            sf_warehouse="test_warehouse",
            sf_storage_integration="test_storage_integration",
            bq_dataset="test_dataset",
            gcs_url="gs://test-location",
            date="2020-01-01",
            overwrite=True
        )
    state = f.run()
    assert state.is_successful()
    mock_cursor.execute.assert_has_calls(
        [
            mock.call("\n        SELECT 1 FROM test_database.test_schema.test_table\n        WHERE session:date='2020-01-01'\n            AND ga_view_id='test_dataset'\n        "), # noqa
            mock.call('\n        CREATE TABLE IF NOT EXISTS test_database.test_schema.test_table (\n            id number autoincrement start 1 increment 1,\n            load_time timestamp_ltz default current_timestamp(),\n            ga_view_id string,\n            session VARIANT\n        );\n        '), # noqa
            mock.call("\n            DELETE FROM test_database.test_schema.test_table\n            WHERE session:date='2020-01-01'\n                AND ga_view_id='test_dataset'\n            "), # noqa
            mock.call("\n        CREATE OR REPLACE STAGE test_database.test_schema.test_table_stage\n            URL = 'gcs://test-location'\n            STORAGE_INTEGRATION = test_storage_integration\n            FILE_FORMAT = (TYPE = JSON);\n        "), # noqa
            mock.call("\n        COPY INTO test_database.test_schema.test_table (ga_view_id, session)\n            FROM (\n                SELECT\n                    'test_dataset',\n                    t.$1\n                FROM @test_database.test_schema.test_table_stage t\n            )\n        PATTERN='.*'\n        FORCE=True\n        "), # noqa
        ]
    )


def test_load_json_objects_to_snowflake_table_exists_no_overwrite(mock_sf_connection):  # noqa: F811
    # Mock the Snowflake connection, cursor, and fetchone method.
    mock_cursor = mock_sf_connection.cursor()
    mock_fetchone = mock.Mock()
    mock_cursor.fetchone = mock_fetchone

    with Flow("test") as f:
        snowflake.load_ga_data_to_snowflake(
            sf_credentials={},
            sf_database="test_database",
            sf_schema="test_schema",
            sf_table="test_table",
            sf_role="test_role",
            sf_warehouse="test_warehouse",
            sf_storage_integration="test_storage_integration",
            bq_dataset="test_dataset",
            gcs_url="gs://test-location",
            date="2020-01-01",
        )
    state = f.run()
    assert state.is_successful()
    mock_cursor.execute.assert_called_once_with("\n        SELECT 1 FROM test_database.test_schema.test_table\n        WHERE session:date='2020-01-01'\n            AND ga_view_id='test_dataset'\n        ") # noqa


def test_load_json_objects_to_snowflake_table_general_exception(mock_sf_connection):
    # Mock the Snowflake connection, cursor, and fetchone method.
    mock_commit = mock.Mock(side_effect=Exception)
    mock_sf_connection.commit = mock_commit

    with Flow("test") as f:
        snowflake.load_ga_data_to_snowflake(
            sf_credentials={},
            sf_database="test_database",
            sf_schema="test_schema",
            sf_table="test_table",
            sf_role="test_role",
            sf_warehouse="test_warehouse",
            sf_storage_integration="test_storage_integration",
            bq_dataset="test_dataset",
            gcs_url="gs://test-location",
            date="2020-01-01",
            overwrite=True
        )
    with raise_on_exception():
        with pytest.raises(Exception):
            f.run()


def test_load_s3_data_to_snowflake_missing_parameters():
    task = snowflake.load_s3_data_to_snowflake
    with pytest.raises(signals.FAIL, match="Either `file` or `pattern` must be specified to run this task."):
        task.run(
            date="2020-01-01",
            date_property='date',
            sf_credentials={},
            sf_database="test_database",
            sf_schema="test_schema",
            sf_table="test_table",
            sf_role="test_role",
            sf_warehouse="test_warehouse",
            sf_storage_integration_name="test_storage_integration",
            s3_url="s3://edx-test/test/",
        )


def test_load_s3_data_to_snowflake_no_existing_table(mock_sf_connection):
    # Mock the Snowflake connection, cursor, and fetchone method.
    mock_cursor = mock_sf_connection.cursor()
    mock_fetchone = mock.Mock(side_effect=ProgrammingError("does not exist"))
    mock_cursor.fetchone = mock_fetchone

    task = snowflake.load_s3_data_to_snowflake
    task.run(
        date="2020-01-01",
        date_property='date',
        sf_credentials={},
        sf_database="test_database",
        sf_schema="test_schema",
        sf_table="test_table",
        sf_role="test_role",
        sf_warehouse="test_warehouse",
        sf_storage_integration_name="test_storage_integration",
        s3_url="s3://edx-test/test/",
        file="test_file.csv",
        pattern=".*",
    )
    mock_cursor.execute.assert_has_calls(
        [
            mock.call("\n            SELECT 1 FROM test_database.test_schema.test_table\n            WHERE date(PROPERTIES:date)=date('2020-01-01')\n            "),  # noqa
            mock.call('\n        CREATE TABLE IF NOT EXISTS test_database.test_schema.test_table (\n            ID NUMBER AUTOINCREMENT START 1 INCREMENT 1,\n            LOAD_TIME TIMESTAMP_LTZ DEFAULT CURRENT_TIMESTAMP(),\n            ORIGIN_FILE_NAME VARCHAR(16777216),\n            ORIGIN_FILE_LINE NUMBER(38,0),\n            ORIGIN_STR VARCHAR(16777216),\n            PROPERTIES VARIANT\n        );\n        '),  # noqa
            mock.call("\n        CREATE STAGE IF NOT EXISTS test_database.test_schema.test_table_stage\n            URL = 's3://edx-test/test/'\n            STORAGE_INTEGRATION = test_storage_integration\n            FILE_FORMAT = (TYPE='JSON', STRIP_OUTER_ARRAY=TRUE);\n        "),  # noqa
            mock.call("\n        COPY INTO test_database.test_schema.test_table (origin_file_name, origin_file_line, origin_str, properties)\n            FROM (\n                SELECT\n                    metadata$filename,\n                    metadata$file_row_number,\n                    t.$1,\n                    CASE\n                        WHEN CHECK_JSON(t.$1) IS NULL THEN t.$1\n                        ELSE NULL\n                    END\n                FROM @test_database.test_schema.test_table_stage t\n            )\n        FILES = ( 'test_file.csv' )\n        PATTERN = '.*'\n        FORCE=False\n        ")  # noqa
        ]
    )


def test_load_s3_data_to_snowflake_data_exists_no_overwrite(mock_sf_connection):
    mock_cursor = mock_sf_connection.cursor()
    mock_fetchone = mock.Mock()
    mock_cursor.fetchone = mock_fetchone

    task = snowflake.load_s3_data_to_snowflake
    with pytest.raises(signals.SKIP, match="Skipping task as data for the date exists and no overwrite was provided."):
        task.run(
            date="2020-01-01",
            date_property='date',
            sf_credentials={},
            sf_database="test_database",
            sf_schema="test_schema",
            sf_table="test_table",
            sf_role="test_role",
            sf_warehouse="test_warehouse",
            sf_storage_integration_name="test_storage_integration",
            s3_url="s3://edx-test/test/",
            pattern=".*",
        )


def test_export_snowflake_table_to_s3_with_exception(mock_sf_connection):
    mock_cursor = mock_sf_connection.cursor()
    mock_execute = mock.Mock(side_effect=ProgrammingError('Files already existing at the unload destination'))
    mock_cursor.execute = mock_execute

    task = snowflake.export_snowflake_table_to_s3
    with pytest.raises(signals.FAIL, match="Files already exist. Use overwrite option to force unloading."):
        task.run(
            sf_credentials={},
            sf_database="test_database",
            sf_schema="test_schema",
            sf_table="test_table",
            sf_role="test_role",
            sf_warehouse="test_warehouse",
            sf_storage_integration="test_storage_integration",
            s3_path="s3://edx-test/test/",
            overwrite=False,
        )


def test_export_snowflake_table_to_s3_overwrite(mock_sf_connection):  # noqa: F811
    mock_cursor = mock_sf_connection.cursor()
    with mock.patch('edx_prefectutils.s3.delete_s3_directory.run') as mock_delete_s3_directory:
        with Flow("test") as f:
            snowflake.export_snowflake_table_to_s3(
                sf_credentials={},
                sf_database="test_database",
                sf_schema="test_schema",
                sf_table="test_table",
                sf_role="test_role",
                sf_warehouse="test_warehouse",
                sf_storage_integration="test_storage_integration",
                s3_path="s3://edx-test/test/",
                overwrite=True,
                enclosed_by='NONE',
                escape_unenclosed_field='\\\\',
                null_marker='NULL',
            )
        state = f.run()
        assert state.is_successful()

        mock_cursor.execute.assert_has_calls(
            [
                mock.call("\n        COPY INTO 's3://edx-test/test/test_database-test_schema-test_table/'\n            FROM test_database.test_schema.test_table\n            STORAGE_INTEGRATION = test_storage_integration\n            FILE_FORMAT = ( TYPE = CSV EMPTY_FIELD_AS_NULL = FALSE\n            FIELD_DELIMITER = ',' FIELD_OPTIONALLY_ENCLOSED_BY = NONE\n            ESCAPE_UNENCLOSED_FIELD = '\\\\'\n            NULL_IF = ( 'NULL' )\n            \n            COMPRESSION = NONE\n            )\n            OVERWRITE=True\n            SINGLE=False\n            DETAILED_OUTPUT = TRUE\n            MAX_FILE_SIZE = 104857600\n    "),  # noqa
            ]
        )

        mock_delete_s3_directory.assert_called_once_with('edx-test', 'test/test_database-test_schema-test_table/')


def test_export_snowflake_table_to_s3_no_escape(mock_sf_connection):  # noqa: F811
    mock_cursor = mock_sf_connection.cursor()
    with mock.patch('edx_prefectutils.s3.delete_s3_directory.run'):
        with Flow("test") as f:
            snowflake.export_snowflake_table_to_s3(
                sf_credentials={},
                sf_database="test_database",
                sf_schema="test_schema",
                sf_table="test_table",
                sf_role="test_role",
                sf_warehouse="test_warehouse",
                sf_storage_integration="test_storage_integration",
                s3_path="s3://edx-test/test/",
                overwrite=True,
                enclosed_by='NONE',
                null_marker='NULL',
            )
        state = f.run()
        assert state.is_successful()

        mock_cursor.execute.assert_has_calls(
            [
                mock.call("\n        COPY INTO 's3://edx-test/test/test_database-test_schema-test_table/'\n            FROM test_database.test_schema.test_table\n            STORAGE_INTEGRATION = test_storage_integration\n            FILE_FORMAT = ( TYPE = CSV EMPTY_FIELD_AS_NULL = FALSE\n            FIELD_DELIMITER = ',' FIELD_OPTIONALLY_ENCLOSED_BY = NONE\n            \n            NULL_IF = ( 'NULL' )\n            \n            COMPRESSION = NONE\n            )\n            OVERWRITE=True\n            SINGLE=False\n            DETAILED_OUTPUT = TRUE\n            MAX_FILE_SIZE = 104857600\n    "),  # noqa
            ]
        )


def test_export_snowflake_table_to_s3_no_enclosure(mock_sf_connection):  # noqa: F811
    mock_cursor = mock_sf_connection.cursor()
    with mock.patch('edx_prefectutils.s3.delete_s3_directory.run'):
        with Flow("test") as f:
            snowflake.export_snowflake_table_to_s3(
                sf_credentials={},
                sf_database="test_database",
                sf_schema="test_schema",
                sf_table="test_table",
                sf_role="test_role",
                sf_warehouse="test_warehouse",
                sf_storage_integration="test_storage_integration",
                s3_path="s3://edx-test/test/",
                overwrite=True,
                escape_unenclosed_field='\\\\',
                null_marker='NULL',
            )
        state = f.run()
        assert state.is_successful()

        mock_cursor.execute.assert_has_calls(
            [
                mock.call("\n        COPY INTO 's3://edx-test/test/test_database-test_schema-test_table/'\n            FROM test_database.test_schema.test_table\n            STORAGE_INTEGRATION = test_storage_integration\n            FILE_FORMAT = ( TYPE = CSV EMPTY_FIELD_AS_NULL = FALSE\n            FIELD_DELIMITER = ',' \n            ESCAPE_UNENCLOSED_FIELD = '\\\\'\n            NULL_IF = ( 'NULL' )\n            \n            COMPRESSION = NONE\n            )\n            OVERWRITE=True\n            SINGLE=False\n            DETAILED_OUTPUT = TRUE\n            MAX_FILE_SIZE = 104857600\n    "),  # noqa
            ]
        )


def test_export_snowflake_table_to_s3_no_null_if(mock_sf_connection):  # noqa: F811
    mock_cursor = mock_sf_connection.cursor()
    with mock.patch('edx_prefectutils.s3.delete_s3_directory.run'):
        with Flow("test") as f:
            snowflake.export_snowflake_table_to_s3(
                sf_credentials={},
                sf_database="test_database",
                sf_schema="test_schema",
                sf_table="test_table",
                sf_role="test_role",
                sf_warehouse="test_warehouse",
                sf_storage_integration="test_storage_integration",
                s3_path="s3://edx-test/test/",
                overwrite=True,
                enclosed_by='NONE',
                escape_unenclosed_field='\\\\',
            )
        state = f.run()
        assert state.is_successful()

        mock_cursor.execute.assert_has_calls(
            [
                mock.call("\n        COPY INTO 's3://edx-test/test/test_database-test_schema-test_table/'\n            FROM test_database.test_schema.test_table\n            STORAGE_INTEGRATION = test_storage_integration\n            FILE_FORMAT = ( TYPE = CSV EMPTY_FIELD_AS_NULL = FALSE\n            FIELD_DELIMITER = ',' FIELD_OPTIONALLY_ENCLOSED_BY = NONE\n            ESCAPE_UNENCLOSED_FIELD = '\\\\'\n            \n            \n            COMPRESSION = NONE\n            )\n            OVERWRITE=True\n            SINGLE=False\n            DETAILED_OUTPUT = TRUE\n            MAX_FILE_SIZE = 104857600\n    "),  # noqa
            ]
        )


def test_export_snowflake_table_to_s3_with_manifest(mock_sf_connection):  # noqa: F811
    mock_cursor = mock_sf_connection.cursor()
    mock_fetchall = mock.Mock()
    s3_files = ['data_0_0_0.csv', 'data_0_0_1.csv']
    mock_fetchall.return_value = [[file] for file in s3_files]
    mock_cursor.fetchall = mock_fetchall

    with mock.patch('prefect.tasks.aws.s3.S3Upload.run') as mock_s3_upload:
        with Flow("test") as f:
            snowflake.export_snowflake_table_to_s3(
                sf_credentials={},
                sf_database="test_database",
                sf_schema="test_schema",
                sf_table="test_table",
                sf_role="test_role",
                sf_warehouse="test_warehouse",
                sf_storage_integration="test_storage_integration",
                s3_path="s3://edx-test/test/",
                overwrite=False,
                generate_manifest=True,
            )
        state = f.run()
        assert state.is_successful()

        expected_manifest_content = {
            "entries": [
                {"url": "s3://edx-test/test/test_database-test_schema-test_table/" + s3_file, "mandatory": True} for s3_file in s3_files # noqa
            ]
        }
        mock_s3_upload.assert_called_once_with(
            json.dumps(expected_manifest_content), key="test/test_database-test_schema-test_table/manifest.json"
        )


def test_export_snowflake_table_to_s3_no_overwrite(mock_sf_connection):  # noqa: F811
    mock_cursor = mock_sf_connection.cursor()

    with Flow("test") as f:
        snowflake.export_snowflake_table_to_s3(
            sf_credentials={},
            sf_database="test_database",
            sf_schema="test_schema",
            sf_table="test_table",
            sf_role="test_role",
            sf_warehouse="test_warehouse",
            sf_storage_integration="test_storage_integration",
            s3_path="s3://edx-test/test/",
            overwrite=False,
            enclosed_by='"',
            escape_unenclosed_field='\\\\',
            null_marker='NULL',
        )
    state = f.run()
    assert state.is_successful()

    mock_cursor.execute.assert_has_calls(
        [
            mock.call("""\n        COPY INTO 's3://edx-test/test/test_database-test_schema-test_table/'\n            FROM test_database.test_schema.test_table\n            STORAGE_INTEGRATION = test_storage_integration\n            FILE_FORMAT = ( TYPE = CSV EMPTY_FIELD_AS_NULL = FALSE\n            FIELD_DELIMITER = ',' FIELD_OPTIONALLY_ENCLOSED_BY = '"'\n            ESCAPE_UNENCLOSED_FIELD = '\\\\'\n            NULL_IF = ( 'NULL' )\n            \n            COMPRESSION = NONE\n            )\n            OVERWRITE=False\n            SINGLE=False\n            DETAILED_OUTPUT = TRUE\n            MAX_FILE_SIZE = 104857600\n    """),  # noqa
        ]
    )


def test_export_snowflake_table_to_s3_with_binary_format(mock_sf_connection):  # noqa: F811
    mock_cursor = mock_sf_connection.cursor()

    with Flow("test") as f:
        snowflake.export_snowflake_table_to_s3(
            sf_credentials={},
            sf_database="test_database",
            sf_schema="test_schema",
            sf_table="test_table",
            sf_role="test_role",
            sf_warehouse="test_warehouse",
            sf_storage_integration="test_storage_integration",
            s3_path="s3://edx-test/test/",
            overwrite=False,
            enclosed_by='"',
            escape_unenclosed_field='\\\\',
            null_marker='NULL',
            binary_format='UTF8',
        )
    state = f.run()
    assert state.is_successful()

    mock_cursor.execute.assert_has_calls(
        [
            mock.call("""\n        COPY INTO 's3://edx-test/test/test_database-test_schema-test_table/'\n            FROM test_database.test_schema.test_table\n            STORAGE_INTEGRATION = test_storage_integration\n            FILE_FORMAT = ( TYPE = CSV EMPTY_FIELD_AS_NULL = FALSE\n            FIELD_DELIMITER = ',' FIELD_OPTIONALLY_ENCLOSED_BY = '"'\n            ESCAPE_UNENCLOSED_FIELD = '\\\\'\n            NULL_IF = ( 'NULL' )\n            BINARY_FORMAT = UTF8\n            COMPRESSION = NONE\n            )\n            OVERWRITE=False\n            SINGLE=False\n            DETAILED_OUTPUT = TRUE\n            MAX_FILE_SIZE = 104857600\n    """),  # noqa
        ]
    )


def test_load_s3_data_to_snowflake_data_disable_check(mock_sf_connection):
    mock_cursor = mock_sf_connection.cursor()
    mock_fetchone = mock.Mock()
    mock_cursor.fetchone = mock_fetchone

    task = snowflake.load_s3_data_to_snowflake
    task.run(
        date="2020-01-01",
        date_property='date',
        sf_credentials={},
        sf_database="test_database",
        sf_schema="test_schema",
        sf_table="test_table",
        sf_role="test_role",
        sf_warehouse="test_warehouse",
        sf_storage_integration_name="test_storage_integration",
        s3_url="s3://edx-test/test/",
        file="test_file.csv",
        pattern=".*",
        overwrite=True,
        disable_existence_check=True,
    )
    mock_call = mock.call("\n            SELECT 1 FROM test_database.test_schema.test_table\n            WHERE date(PROPERTIES:date)=date('2020-01-01')\n            ")  # noqa

    assert mock_call not in mock_cursor.execute.mock_calls

    task.run(
        date="2020-01-01",
        date_property='date',
        sf_credentials={},
        sf_database="test_database",
        sf_schema="test_schema",
        sf_table="test_table",
        sf_role="test_role",
        sf_warehouse="test_warehouse",
        sf_storage_integration_name="test_storage_integration",
        s3_url="s3://edx-test/test/",
        file="test_file.csv",
        pattern=".*",
        overwrite=True,
        disable_existence_check=False,
    )
    mock_cursor.execute.assert_has_calls(
        [
            mock_call
        ]
    )



"""Test the API client"""

import json
from datetime import datetime, timedelta
from unittest import TestCase

import httpretty
import requests
from ddt import data, ddt, unpack
from mock import patch

from edx_prefectutils.edx_api_client import EdxApiClient

FAKE_AUTH_URL = 'http://example.com/oauth2/access_token'
FAKE_CLIENT_ID = 'aclientid'
FAKE_CLIENT_SECRET = 'notasecret'
FAKE_ACCESS_TOKEN = 'notasecrettoken'
FAKE_RESOURCE_URL = 'http://example.com/resource/'


@ddt
@httpretty.activate
class EdxApiClientTestCase(TestCase):
    """Test the client"""

    def setUp(self):
        self.current_time = datetime(2016, 1, 1, 0, 0, 0, 0)
        self.time_offset = 0

        class MockDateTime(datetime):
            """
            A fake datetime that allows the test to control the return value from the utcnow() method.

            This allows us to emulate the passage of time, by simply changing `self.time_offset`.
            """

            @classmethod
            def utcnow(cls):
                """Return the time specified by the time offset"""
                return self.current_time + timedelta(seconds=self.time_offset)

        datetime_patcher = patch('edx_prefectutils.edx_api_client.datetime', MockDateTime)
        datetime_patcher.start()
        self.addCleanup(datetime_patcher.stop)

        self.client = EdxApiClient(auth_url=FAKE_AUTH_URL, client_id=FAKE_CLIENT_ID, client_secret=FAKE_CLIENT_SECRET)

    def test_fetch_access_token(self):
        self.prepare_for_token_request()

        self.client.ensure_oauth_access_token()

        self.assertEqual(self.client.authenticated_session.auth.token, FAKE_ACCESS_TOKEN)

        self.assert_token_request_body(FAKE_CLIENT_ID, FAKE_CLIENT_SECRET)

    def assert_token_request_body(self, client_id, client_secret, token_type='jwt'):
        """Look at the most recent request and make sure it contained the provided credentials."""
        token_request = httpretty.last_request()
        self.assertEqual(token_request.method, 'POST')
        self.assertEqual(
            token_request.parsed_body,
            {
                'token_type': [token_type],
                'client_secret': [client_secret],
                'client_id': [client_id],
                'grant_type': ['client_credentials']
            }
        )

    @staticmethod
    def prepare_for_token_request(expires_in=60, access_token=FAKE_ACCESS_TOKEN):
        """Setup the request mock so that it returns the specified response."""
        response = {
            'expires_in': expires_in,
            'access_token': access_token
        }
        httpretty.register_uri(httpretty.POST, FAKE_AUTH_URL, body=json.dumps(response))

    def test_secrets_from_config(self):
        self.prepare_for_token_request()

        client = EdxApiClient(client_id='cidfromcfg', client_secret='secfromcfg', auth_url=FAKE_AUTH_URL)
        client.ensure_oauth_access_token()

        self.assert_token_request_body('cidfromcfg', 'secfromcfg')
        self.assertEqual(client.authenticated_session.auth.token, FAKE_ACCESS_TOKEN)

    def test_token_type(self):
        self.prepare_for_token_request()

        client = EdxApiClient(
            client_id='cidfromcfg',
            client_secret='secfromcfg',
            auth_url=FAKE_AUTH_URL,
            token_type='bearer',
        )
        client.ensure_oauth_access_token()

        self.assert_token_request_body('cidfromcfg', 'secfromcfg', token_type='bearer')
        self.assertEqual(client.authenticated_session.auth.token, FAKE_ACCESS_TOKEN)

    def test_expiration(self):
        expires_in = 10
        self.prepare_for_token_request(expires_in=expires_in, access_token='token1')
        self.client.ensure_oauth_access_token()

        self.assertEqual(self.client.authenticated_session.auth.token, 'token1')

        self.time_offset = 11
        self.prepare_for_token_request(expires_in=10, access_token='token2')
        self.client.ensure_oauth_access_token()

        self.assertEqual(self.client.authenticated_session.auth.token, 'token2')

    def test_get_single_page(self):
        self.prepare_for_token_request()
        response_body = {
            'results': [
                {'a': 1},
                {'a': 2}
            ]
        }
        httpretty.register_uri('GET', FAKE_RESOURCE_URL, body=json.dumps(response_body))

        response = self.client.get(FAKE_RESOURCE_URL)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), response_body)

        responses = list(self.client.paginated_get(FAKE_RESOURCE_URL))
        self.assertEqual(len(responses), 1)
        self.assertEqual(response.status_code, responses[0].status_code)
        self.assertEqual(response.json(), responses[0].json())

    def test_get_internal_server_error(self):
        self.prepare_for_token_request()
        httpretty.register_uri('GET', FAKE_RESOURCE_URL, body='Server Error', status=500)

        with self.assertRaises(requests.HTTPError):
            list(self.client.paginated_get(FAKE_RESOURCE_URL))

    @patch('time.sleep', sleep_mock=None)
    def test_non_fatal_error(self, sleep_mock):
        self.prepare_for_token_request()
        httpretty.register_uri('GET', FAKE_RESOURCE_URL,
                               responses=[
                                   httpretty.Response(body='Too many requests!', status=429),
                                   httpretty.Response(body='Unavailable', status=503),
                                   httpretty.Response(body='{}', status=200)
                               ])

        responses = list(self.client.paginated_get(FAKE_RESOURCE_URL))
        self.assertEqual(len(responses), 1)
        response = responses[0]
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {})

        num_auth_token_requests = 1
        num_failed_requests = 2
        num_successful_requests = 1
        total_expected_requests = num_auth_token_requests + num_failed_requests + num_successful_requests
        self.assertEqual(len(httpretty.httpretty.latest_requests), total_expected_requests)

    @data(
        ([
            {
                'next': FAKE_RESOURCE_URL + '?offset=2&limit=2&foo=bar',
                'results': [{'a': 1}, {'a': 2}]
            },
            {
                'next': FAKE_RESOURCE_URL + '?offset=4&limit=2&foo=bar',
                'results': [{'a': 3}, {'a': 4}]
            },
            {
                'next': None,
                'results': [{'a': 5}, {'a': 6}]
            }
        ], 'next'),
        ([
            {
                'pagination': {'next': FAKE_RESOURCE_URL + '?offset=2&limit=2&foo=bar'},
                'results': [{'a': 1}, {'a': 2}]
            },
            {
                'pagination': {'next': FAKE_RESOURCE_URL + '?offset=4&limit=2&foo=bar'},
                'results': [{'a': 3}, {'a': 4}]
            },
            {
                'results': [{'a': 5}, {'a': 6}]
            }
        ], lambda r: r.get('pagination', {}).get('next'))
    )
    @unpack
    @patch('time.sleep', sleep_mock=None)
    def test_paginated_get(self, response_bodies, pagination_key, sleep_mock):
        self.prepare_for_token_request()
        httpretty.register_uri('GET', FAKE_RESOURCE_URL,
                               responses=[
                                   httpretty.Response(body=json.dumps(response_bodies[0])),
                                   httpretty.Response(body='Unavailable', status=503),
                                   httpretty.Response(body=json.dumps(response_bodies[1])),
                                   httpretty.Response(body='Too many requests!', status=429),
                                   httpretty.Response(body=json.dumps(response_bodies[2])),
                               ])

        responses = list(self.client.paginated_get(FAKE_RESOURCE_URL, params={'limit': 2, 'foo': 'bar'},
                                                   pagination_key=pagination_key))
        self.assertEqual([response.json() for response in responses], response_bodies)

        self.assertEqual(len(httpretty.httpretty.latest_requests), 6)
        self.assertEqual(
            httpretty.httpretty.latest_requests[1].querystring, {'limit': ['2'], 'foo': ['bar']}
        )
        self.assertEqual(
            httpretty.httpretty.latest_requests[3].querystring, {'limit': ['2'], 'foo': ['bar'], 'offset': ['2']}
        )
        self.assertEqual(
            httpretty.httpretty.latest_requests[5].querystring, {'limit': ['2'], 'foo': ['bar'], 'offset': ['4']}
        )



"""
Utility methods and tasks for working with Snowflake from a Prefect flow.
"""
import json
import os
from collections import namedtuple
from typing import List, TypedDict
from urllib.parse import urlparse

import backoff
import snowflake.connector
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from prefect import task
from prefect.engine import signals
from prefect.tasks.aws import s3
from prefect.utilities.logging import get_logger

from edx_prefectutils import s3 as s3_utils

MANIFEST_FILE_NAME = 'manifest.json'
EXPORT_MAX_FILESIZE = 104857600


class SFCredentials(TypedDict, total=False):
    private_key: str
    private_key_passphrase: str
    user: str
    account: str
    password: str


def create_snowflake_connection(
    credentials: SFCredentials,
    role: str,
    autocommit: bool = False,
    warehouse: str = None
) -> snowflake.connector.SnowflakeConnection:
    """
    Connects to the Snowflake database.

      credentials (SFCredentials):
        Snowflake credentials including key & passphrase, along with user and account.
      role (str): Name of the role to use for the connection.
      autocommit (bool): True to enable autocommit for the connection, False if not.
      warehouse (str): The Snowflake warehouse to use for this connection. Defaults to the user's default warehouse.
    """
    private_key = credentials.get("private_key")

    private_key_passphrase = credentials.get("private_key_passphrase")
    user = credentials.get("user")
    account = credentials.get("account")
    password = credentials.get("password")

    p_key = None
    pkb = None
    if private_key and private_key_passphrase:
        p_key = serialization.load_pem_private_key(
            private_key.encode(),
            password=private_key_passphrase.encode(),
            backend=default_backend(),
        )

        pkb = p_key.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

    connection = snowflake.connector.connect(
        user=user, account=account, private_key=pkb, password=password, autocommit=autocommit, warehouse=warehouse
    )

    # Switch to specified role.
    connection.cursor().execute("USE ROLE {}".format(role))
    # Set timezone to UTC
    connection.cursor().execute("ALTER SESSION SET TIMEZONE = 'UTC'")

    return connection


def qualified_table_name(database, schema, table) -> str:
    """
    Fully qualified Snowflake table name.
    """
    return "{database}.{schema}.{table}".format(
        database=database, schema=schema, table=table
    )


def qualified_stage_name(database, schema, table) -> str:
    """
    Fully qualified Snowflake stage name.
    """
    return "{database}.{schema}.{table}_stage".format(
        database=database, schema=schema, table=table,
    )


@task
@backoff.on_exception(backoff.expo,
                      snowflake.connector.ProgrammingError,
                      max_tries=3)
def load_ga_data_to_snowflake(
    sf_credentials: SFCredentials,
    sf_database: str,
    sf_schema: str,
    sf_table: str,
    sf_role: str,
    sf_warehouse: str,
    sf_storage_integration: str,
    bq_dataset: str,
    gcs_url: str,
    date: str,
    pattern: str = ".*",
    overwrite: bool = False,
):
    """
    Loads JSON objects from GCS to Snowflake.
    Args:
      sf_credentials (SFCredentials):
        Snowflake public key credentials in the format required by create_snowflake_connection.
      sf_database (str): Name of the destination database.
      sf_schema (str): Name of the destination schema.
      sf_table (str): Name of the destination table.
      sf_role (str): Name of the snowflake role to assume.
      sf_warehouse (str): Name of the Snowflake warehouse to be used for loading.
      sf_storage_integration (str):
        The name of the pre-configured storage integration created for this flow.
      bq_dataset (str): BQ Dataset to which this load belongs to. This gets set as `ga_view_id' in the dest. table.
      gcs_url (str): Full URL to the GCS path containing the files to load.
      pattern (str, optional): Path pattern/regex to match GCS object to copy.
      date (str): Date of `ga_sessions` being loaded.
      overwrite (bool, optional): Whether to overwrite existing data for the given date. Defaults to `False`.
    """
    sf_connection = create_snowflake_connection(sf_credentials, sf_role)
    # Snowflake expects GCS locations to start with `gcs` instead of `gs`.
    gcs_url = gcs_url.replace("gs://", "gcs://")

    # Check for data existence for this date
    try:
        query = """
        SELECT 1 FROM {table}
        WHERE session:date='{date}'
            AND ga_view_id='{ga_view_id}'
        """.format(
            table=qualified_table_name(sf_database, sf_schema, sf_table),
            date=date,
            ga_view_id=bq_dataset,
        )
        cursor = sf_connection.cursor()
        cursor.execute(query)
        row = cursor.fetchone()
    except snowflake.connector.ProgrammingError as e:
        if "does not exist" in e.msg:
            # If so then the query failed because the table doesn't exist.
            row = None
        else:
            raise

    if row and not overwrite:
        return

    try:
        query = """
        CREATE TABLE IF NOT EXISTS {table} (
            id number autoincrement start 1 increment 1,
            load_time timestamp_ltz default current_timestamp(),
            ga_view_id string,
            session VARIANT
        );
        """.format(
            table=qualified_table_name(sf_database, sf_schema, sf_table)
        )
        sf_connection.cursor().execute(query)

        if overwrite:
            query = """
            DELETE FROM {table}
            WHERE session:date='{date}'
                AND ga_view_id='{ga_view_id}'
            """.format(
                table=qualified_table_name(sf_database, sf_schema, sf_table),
                date=date,
                ga_view_id=bq_dataset,
            )
            sf_connection.cursor().execute(query)

        query = """
        CREATE OR REPLACE STAGE {stage_name}
            URL = '{stage_url}'
            STORAGE_INTEGRATION = {storage_integration}
            FILE_FORMAT = (TYPE = JSON);
        """.format(
            stage_name=qualified_stage_name(sf_database, sf_schema, sf_table),
            stage_url=gcs_url,
            storage_integration=sf_storage_integration,
        )
        sf_connection.cursor().execute(query)

        query = """
        COPY INTO {table} (ga_view_id, session)
            FROM (
                SELECT
                    '{ga_view_id}',
                    t.$1
                FROM @{stage_name} t
            )
        PATTERN='{pattern}'
        FORCE={force}
        """.format(
            table=qualified_table_name(sf_database, sf_schema, sf_table),
            ga_view_id=bq_dataset,
            stage_name=qualified_stage_name(sf_database, sf_schema, sf_table),
            pattern=pattern,
            force=str(overwrite),
        )
        sf_connection.cursor().execute(query)
        sf_connection.commit()
    except Exception:
        sf_connection.rollback()
        raise
    finally:
        sf_connection.close()


@task
def load_s3_data_to_snowflake(
    date: str,
    date_property: str,
    sf_credentials: SFCredentials,
    sf_database: str,
    sf_schema: str,
    sf_table: str,
    sf_role: str,
    sf_warehouse: str,
    sf_storage_integration_name: str,
    s3_url: str,
    sf_file_format: str = "TYPE='JSON', STRIP_OUTER_ARRAY=TRUE",
    file: str = None,
    pattern: str = None,
    overwrite: bool = False,
    truncate: bool = False,
    disable_existence_check: bool = False,
):
    """
    Loads objects in S3 to a generic table in Snowflake, the data is stored in a variant column named
    `PROPERTIES`. Note that either `file` or `pattern` parameter must be specified.
    Notes:
      To load a single file, use the `file` parameter.
      You must explicitly include a separator (/) either at the end of the `s3_url` or at the beginning
      of file path specified in the `file` parameter. Typically you would include a trailing (/)
      in `s3_url`.
      To load multiple files, use `pattern`.
    Examples:
      To load a single file `s3://bucket/path/to/filename/filename.ext`:
      load_s3_data_to_snowflake(s3_url='s3://bucket/path/to/filename/, file='filename.ext')
      or
      load_s3_data_to_snowflake(s3_url='s3://bucket/path/, file='to/filename/filename.ext')

      Load all files under `s3://bucket/path/` with a certain name:
      load_s3_data_to_snowflake(s3_url='s3://bucket/path/, pattern='.*filename.ext')

    Args:
      date (str): Date of the data being loaded.
      date_property (str): Date type property name in the variant `PROPERTIES` column.
      sf_credentials (SFCredentials): Snowflake public key credentials in the
              format required by create_snowflake_connection.
      sf_database (str): Name of the destination database.
      sf_schema (str): Name of the destination schema.
      sf_table (str): Name of the destination table.
      sf_role (str): Name of the snowflake role to assume.
      sf_warehouse (str): Name of the Snowflake warehouse to be used for loading.
      sf_storage_integration_name (str): Name of the Snowflake storage integration to use. These are
              configured in Terraform.
      s3_url (str): Full URL to the S3 path containing the files to load.
      sf_file_format (str, optional): Snowflake file format for the Stage. Defaults to 'JSON'.
      file (str, optional): File path relative to `s3_url`.
      pattern (str, optional): Path pattern/regex to match S3 objects to copy. Defaults to `None`.
      overwrite (bool, optional): Whether to overwrite existing data for the given date. Defaults to `False`.
      truncate (bool, optional): Whether to truncate the table. Defaults to `False`.
      disable_existence_check (bool, optional): Whether to disable check for existing data, useful when
              always appending to the table regardless of any existing data for that provided `date`
    """
    logger = get_logger()
    if not file and not pattern:
        raise signals.FAIL('Either `file` or `pattern` must be specified to run this task.')

    sf_connection = create_snowflake_connection(sf_credentials, sf_role, warehouse=sf_warehouse)

    if truncate:
        query = "TRUNCATE IF EXISTS {}".format(qualified_table_name(sf_database, sf_schema, sf_table))
        logger.info("Truncating table: {}".format(sf_table))
        cursor = sf_connection.cursor()
        cursor.execute(query)

    # Check for data existence for this date
    row = None
    if not disable_existence_check:
        try:
            query = """
            SELECT 1 FROM {table}
            WHERE date(PROPERTIES:{date_property})=date('{date}')
            """.format(
                table=qualified_table_name(sf_database, sf_schema, sf_table),
                date=date,
                date_property=date_property,
            )

            logger.info("Checking existence of data for {}".format(date))

            cursor = sf_connection.cursor()
            cursor.execute(query)
            row = cursor.fetchone()
        except snowflake.connector.ProgrammingError as e:
            if "does not exist" in e.msg:
                # If so then the query failed because the table doesn't exist.
                row = None
            else:
                raise

    if row and not overwrite:
        raise signals.SKIP('Skipping task as data for the date exists and no overwrite was provided.')
    else:
        logger.info("Continuing with S3 load for {}".format(date))

    try:
        # Create the generic loading table
        query = """
        CREATE TABLE IF NOT EXISTS {table} (
            ID NUMBER AUTOINCREMENT START 1 INCREMENT 1,
            LOAD_TIME TIMESTAMP_LTZ DEFAULT CURRENT_TIMESTAMP(),
            ORIGIN_FILE_NAME VARCHAR(16777216),
            ORIGIN_FILE_LINE NUMBER(38,0),
            ORIGIN_STR VARCHAR(16777216),
            PROPERTIES VARIANT
        );
        """.format(
            table=qualified_table_name(sf_database, sf_schema, sf_table)
        )

        sf_connection.cursor().execute(query)

        # Delete existing data in case of overwrite.
        if overwrite and row:
            logger.info("Deleting data for overwrite for {}".format(date))

            query = """
            DELETE FROM {table}
            WHERE date(PROPERTIES:{date_property})=date('{date}')
            """.format(
                table=qualified_table_name(sf_database, sf_schema, sf_table),
                date=date,
                date_property=date_property,
            )
            sf_connection.cursor().execute(query)

        # Create stage
        query = """
        CREATE STAGE IF NOT EXISTS {stage_name}
            URL = '{stage_url}'
            STORAGE_INTEGRATION = {storage_integration}
            FILE_FORMAT = ({file_format});
        """.format(
            stage_name=qualified_stage_name(sf_database, sf_schema, sf_table),
            stage_url=s3_url,
            storage_integration=sf_storage_integration_name,
            file_format=sf_file_format,
        )
        sf_connection.cursor().execute(query)

        files_paramater = ""
        pattern_parameter = ""

        if file:
            logger.info("Loading file {}".format(file))
            files_paramater = "FILES = ( '{}' )".format(file)

        if pattern:
            logger.info("Loading pattern {}".format(pattern))
            pattern_parameter = "PATTERN = '{}'".format(pattern)

        query = """
        COPY INTO {table} (origin_file_name, origin_file_line, origin_str, properties)
            FROM (
                SELECT
                    metadata$filename,
                    metadata$file_row_number,
                    t.$1,
                    CASE
                        WHEN CHECK_JSON(t.$1) IS NULL THEN t.$1
                        ELSE NULL
                    END
                FROM @{stage_name} t
            )
        {files_parameter}
        {pattern_parameter}
        FORCE={force}
        """.format(
            table=qualified_table_name(sf_database, sf_schema, sf_table),
            stage_name=qualified_stage_name(sf_database, sf_schema, sf_table),
            files_parameter=files_paramater,
            pattern_parameter=pattern_parameter,
            force=str(overwrite),
        )

        logger.info("Copying data into Snowflake as: \n{}".format(query))

        sf_connection.cursor().execute(query)
        sf_connection.commit()
    except Exception:
        sf_connection.rollback()
        raise
    finally:
        sf_connection.close()


@task
def export_snowflake_table_to_s3(
    sf_credentials: SFCredentials,
    sf_database: str,
    sf_schema: str,
    sf_table: str,
    sf_role: str,
    sf_warehouse: str,
    sf_storage_integration: str,
    s3_path: str,
    field_delimiter: str = ',',
    enclosed_by: str = None,
    escape_unenclosed_field: str = None,
    null_marker: str = None,
    binary_format: str = None,
    overwrite: bool = True,
    single: bool = False,
    generate_manifest: bool = False,
):

    """
    Exports a snowflake table to S3.
    The output path is generated by appending fully qualified table name to `s3_path` parameter.

    Args:
      sf_credentials (SFCredentials): Snowflake public key credentials in the format
              required by create_snowflake_connection.
      sf_database (str): Name of the source database.
      sf_schema (str): Name of the source schema.
      sf_table (str): Name of the source table.
      sf_role (str): Name of the snowflake role to assume.
      sf_warehouse (str): Name of the Snowflake warehouse to be used for loading.
      sf_storage_integration_name (str): Name of the Snowflake storage integration to use. These are
              configured in Terraform.
      s3_path (str): S3 base path used to unload the table, this will be appended with the qualified table name.
      field_delimiter (str, optional): The character to use for separating fields in the output file. Defaults to `,`.
      enclosed_by (str, optional): Character used to enclose strings or None to use snowflake default
              'NONE' is a specially allowed value which emits NONE to snowflake, meaning no enclosure
      escape_unenclosed_field (str, optional): Single character string used as the escape character for unenclosed
              field values only. Defaults to None which omits this clause letting snowflake use its default \\.
              'NONE' is a specially allowed value which emits NONE to snowflake, meaning no escaping ever.
      null_marker (str, optional): String used to convert SQL NULL. Defaults to None which lets snowflake
              use its default '\\N'
      binary_format (str, optional): String to define encoding for binary output
      overwrite (bool, optional): Whether to overwrite existing data in S3. Defaults to `TRUE`.
      single (bool, optional): Whether to generate a single file in S3. Defaults to `FALSE`. The maximum file size
              for a single file defaults to 16MB, although that default can be updated by adding a MAX_FILE_SIZE
              copy option.
      generate_manifest (bool, optional): Whether to generate a manifest file in S3. Defaults to `FALSE`.
    """
    logger = get_logger()

    sf_connection = create_snowflake_connection(
        credentials=sf_credentials,
        role=sf_role,
        warehouse=sf_warehouse,
    )

    table_name = qualified_table_name(sf_database, sf_schema, sf_table)
    export_path = os.path.join(s3_path, table_name.replace('.', '-').lower()) + '/'

    parsed_path = urlparse(export_path)
    export_bucket = parsed_path.netloc
    export_prefix = parsed_path.path.lstrip('/')

    # Snowflake's COPY INTO command's OVERWRITE option is not reliable, so we need to delete the data manually
    if overwrite:
        logger.info("Deleting existing data in S3 bucket: {bucket} with prefix: {prefix}".format(
            bucket=export_bucket, prefix=export_prefix))
        s3_utils.delete_s3_directory.run(export_bucket, export_prefix)

    escape_clause = '' if escape_unenclosed_field is None \
        else "ESCAPE_UNENCLOSED_FIELD = NONE" if escape_unenclosed_field == 'NONE' \
        else "ESCAPE_UNENCLOSED_FIELD = '{escape_unenclosed_field}'".format(
            escape_unenclosed_field=escape_unenclosed_field)
    enclosure_clause = '' if enclosed_by is None \
        else "FIELD_OPTIONALLY_ENCLOSED_BY = NONE" if enclosed_by == 'NONE' \
        else "FIELD_OPTIONALLY_ENCLOSED_BY = '{enclosed_by}'".format(enclosed_by=enclosed_by)
    null_if_clause = '' if null_marker is None \
        else "NULL_IF = ( '{null_marker}' )".format(null_marker=null_marker)
    binary_format_clause = '' if binary_format is None \
        else "BINARY_FORMAT = {binary_format}".format(
            binary_format=binary_format)

    query = """
        COPY INTO '{export_path}'
            FROM {table}
            STORAGE_INTEGRATION = {storage_integration}
            FILE_FORMAT = ( TYPE = CSV EMPTY_FIELD_AS_NULL = FALSE
            FIELD_DELIMITER = '{field_delimiter}' {enclosure_clause}
            {escape_clause}
            {null_if_clause}
            {binary_format_clause}
            COMPRESSION = NONE
            )
            OVERWRITE={overwrite}
            SINGLE={single}
            DETAILED_OUTPUT = TRUE
            MAX_FILE_SIZE = {max_file_size}
    """.format(
        export_path=export_path,
        table=table_name,
        storage_integration=sf_storage_integration,
        field_delimiter=field_delimiter,
        enclosure_clause=enclosure_clause,
        escape_clause=escape_clause,
        null_if_clause=null_if_clause,
        binary_format_clause=binary_format_clause,
        overwrite=overwrite,
        single=single,
        max_file_size=EXPORT_MAX_FILESIZE,
    )
    logger.info(query)
    cursor = sf_connection.cursor()

    try:
        cursor.execute(query)
        if generate_manifest:
            # With DETAILED_OUTPUT=TRUE, the COPY command returns rows for each file that gets created.
            # The first column is the FILE_NAME
            s3_file_names = [row[0] for row in cursor.fetchall()]
            # Generate a manifest file in S3 which can be later used by LOAD DATA FROM S3 MANIFEST
            # command to load the data into MySQL.
            s3_file_paths = [os.path.join(export_path, s3_file_name) for s3_file_name in s3_file_names]
            s3_manifest_file_prefix = os.path.join(export_prefix, MANIFEST_FILE_NAME)
            manifest_file_content = {
                "entries": [
                    {"url": s3_file_path, "mandatory": True} for s3_file_path in s3_file_paths
                ]
            }

            logger.info(
                'Uploading manifest file to s3, containing urls %(urls)s',
                {
                    "urls": s3_file_paths,
                }
            )
            s3.S3Upload(bucket=export_bucket).run(
                json.dumps(manifest_file_content),
                key=s3_manifest_file_prefix
            )
    except snowflake.connector.ProgrammingError as e:
        if 'Files already existing at the unload destination' in e.msg:
            logger.error("Files already exist at {destination}".format(destination=export_path))
            raise signals.FAIL('Files already exist. Use overwrite option to force unloading.')
        else:
            raise
    finally:
        sf_connection.close()

    return export_path


def get_batched_rows_from_snowflake(
    sf_credentials: SFCredentials,
    sf_database: str,
    sf_schema: str,
    sf_table: str,
    sf_role: str,
    columns: List[str],
    batch_size: int,
    where: str = None,
):
    """
    Query batches of rows from snowflake as a generator.

    Args:
      sf_credentials (SFCredentials): Snowflake public key credentials in the format
              required by create_snowflake_connection.
      sf_database (str): Name of the source database.
      sf_schema (str): Name of the source schema.
      sf_table (str): Name of the source table.
      sf_role (str): Name of the snowflake role to assume.
      columns (list[str]): The column field names you want to select.
      batch_size (int): How many at most to return at once.
      where (str, optional): A WHERE query string for the snowflake SELECT call. None will return all rows.

    Returns:
      A generator that yields lists of row objects. Each list will be at most `batch_size` long.
      Each row object will be a named tuple with the field names of the provided `columns`.

    Examples:
      for rows in get_batched_rows_from_snowflake(..., ['foo', 'bar'], 75):
        for row in rows:
          print('Got foo %s and bar %s', row.foo, row.bar)

      for rows in get_batched_rows_from_snowflake(..., ['foo'], 50, where='bar is null'):
        for row in rows:
          print('Got foo %s', row.foo)
    """
    sf_connection = create_snowflake_connection(
        sf_credentials,
        sf_role,
    )

    query = 'SELECT {columns} FROM {table}'.format(
        columns=', '.join(columns),
        table=qualified_table_name(sf_database, sf_schema, sf_table),
    )
    if where:
        query += ' WHERE ' + where

    cursor = sf_connection.cursor()
    cursor.execute(query)

    SnowFlakeRow = namedtuple('SnowFlakeRow', columns)
    batch = cursor.fetchmany(batch_size)
    while len(batch) > 0:
        yield [SnowFlakeRow(*row) for row in batch]
        batch = cursor.fetchmany(batch_size)



"""
Utility methods and tasks for use from a Prefect flow.
"""

import datetime
import itertools
import re

import prefect
import six
from opaque_keys import InvalidKeyError
from opaque_keys.edx.keys import CourseKey
from prefect import task
from prefect.engine.results import PrefectResult


@task
def get_date(date: str):
    """
    Return today's date string if date is None. Otherwise return the passed parameter value.
    prefect.context.today is only available at task level, so we cannot use it as a default parameter value.
    """
    if date is None:
        return prefect.context.today
    else:
        return date


@task(result=PrefectResult())
def generate_dates(start_date: str, end_date: str, date_format: str = "%Y%m%d"):
    """
    Generates a list of date strings in the format specified by `date_format` from
    start_date up to but excluding end_date.
    """
    if not start_date:
        start_date = prefect.context.yesterday
    if not end_date:
        end_date = prefect.context.today

    parsed_start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d")
    parsed_end_date = datetime.datetime.strptime(end_date, "%Y-%m-%d")
    dates = []
    while parsed_start_date < parsed_end_date:
        dates.append(parsed_start_date)
        parsed_start_date = parsed_start_date + datetime.timedelta(days=1)

    return [date.strftime(date_format) for date in dates]


@task
def generate_month_start_dates(start_date: str, end_date: str, date_format: str = "%Y-%m-%d"):
    """
    Return a list of first days of months within the specified date range.
    If start_date or end_date is not provided, defaults to yesterday or today respectively.
    prefect.context.today is only available at task level, so we cannot use it as a default parameter value.
    """
    if not start_date:
        start_date = prefect.context.yesterday
    if not end_date:
        end_date = prefect.context.today

    # Since our intention is to extract first day of months, we will start by modifying the start and end date
    # to represent the first day of month.
    parsed_start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d").replace(day=1)
    parsed_end_date = datetime.datetime.strptime(end_date, "%Y-%m-%d").replace(day=1)
    dates = []
    current_date = parsed_start_date
    while current_date <= parsed_end_date:
        dates.append(current_date)
        # The addition of 32 days to current_date and then setting the day to 1 is a way to ensure that we move to
        # the beginning of the next month, even if the month doesn't have exactly 32 days.
        current_date += datetime.timedelta(days=32)
        current_date = current_date.replace(day=1)

    return [date.strftime(date_format) for date in dates]


@task
def get_unzipped_cartesian_product(input_lists: list):
    """
    Generate an unzipped cartesian product of the given list of lists, useful for
    generating task parameters for mapping.

    For example, get_unzipped_cartesian_product([[1, 2, 3], ["a", "b", "c"]]) would return:

    [
      [1, 1, 1, 2, 2, 3, 3, 3],
      ["a", "b", "c", "a", "b", "c", "a", "b", "c"]
    ]

    Args:
      input_lists (list): A list of two or more lists.
    """
    return list(zip(*itertools.product(*input_lists)))


def get_filename_safe_course_id(course_id, replacement_char='_'):
    """
    Create a representation of a course_id that can be used safely in a filepath.
    """
    try:
        course_key = CourseKey.from_string(course_id)
        # Ignore the namespace of the course_id altogether, for backwards compatibility.
        filename = course_key._to_string()  # pylint: disable=protected-access
    except InvalidKeyError:
        # If the course_id doesn't parse, we will still return a value here.
        filename = course_id

    # The safest characters are A-Z, a-z, 0-9, <underscore>, <period> and <hyphen>.
    # We represent the first four with \w.
    # TODO: Once we support courses with unicode characters, we will need to revisit this.
    return re.sub(r'[^\w\.\-]', six.text_type(replacement_char), filename)



import xml.etree.cElementTree as ET
from collections import OrderedDict, namedtuple
from decimal import Decimal
from io import BytesIO

import requests
import six


class PaypalApiResponse(object):
    """
    A generic API response. All API responses include the "baseResponse" element which is parsed by this class so any
    children overriding `params_from_xml` are expected to call this class.

    Subclasses are expected to take in the response_code and response_message in their constructor and pass them up to
    the constructor of this class.

    Args:
        response_code: The numeric status code for the request, this indicates the overall request status.
        response_message: A human readable description of the status of the request.
    """

    SUCCESSFUL_RESPONSE_CODE = 100

    def __init__(self, response_code, response_message):
        # NOTE: subclasses overriding this method should also include these parameters in their signature since the
        # constructor call is generated dynamically during the XML parsing process.
        self.response_code = response_code
        self.response_message = response_message

    @classmethod
    def from_http_response(cls, response):
        """
        Parse an HTTP response from the API. Raises errors if the response is not usable. Returns the populated response
        object if it is.

        Args:
            response: a requests.HTTPResponse object

        Raises:
            HTTPError: a requests.exceptions.HTTPError object if the API returned an HTTP status code other than 200 OK.
            PaypalError: If either the response was malformed or Paypal indicated that it could not process the
                request for some reason.
        """
        response.raise_for_status()
        response_root = ET.fromstring(response.content)
        response_obj = cls.from_xml(response_root)
        response_obj.raise_for_status()
        return response_obj

    @classmethod
    def params_from_xml(cls, _root_node):
        """
        Subclasses are expected to override this method to parse fields out of the XML document to be passed to their
        constructor.

        Args:
            root_node: an ElementTree.Element object representing the root node of the XML response.

        Returns:
            A dictionary mapping constructor arguments to the values extracted from the XML document.
        """
        return {}

    @classmethod
    def from_xml(cls, root_node):
        """
        Parse the status of the request from the response. This request status should be included in all responses.

        Args:
            root_node: an ElementTree.Element object representing the root node of the XML response.

        Returns:
            A valid response object.
        """
        base_response_node = find_or_raise(root_node, 'baseResponse')
        params = {
            'response_code': int(find_text_or_raise(base_response_node, 'responseCode')),
            'response_message': find_text_or_raise(base_response_node, 'responseMsg')
        }
        params.update(cls.params_from_xml(root_node))
        return cls(**params)

    def raise_for_status(self):
        """Raise an error if the request failed or the response is malformed."""
        if self.response_code != self.SUCCESSFUL_RESPONSE_CODE:
            raise PaypalApiRequestFailedError(self.response_code, self.response_message)


def find_or_raise(node, child_name):
    """
    Find the direct descendant of `node` with the tag `child_name`. If the no child is found with that tag, an error is
    raised.

    Args:
        node: an ElementTree.Element object representing a node in an XML document that is expected to have a child
            with the name specified by the `child_name` parameter.
        child_name: a string containing the name of the tag that is expected to be a direct descendant of `node`.

    Raises:
        PaypalMalformedResponseError: If the document does not have the expected structure.

    Returns:
        A reference to the child node when it is found.
    """
    child = node.find(child_name)
    if child is None:
        raise PaypalMalformedResponseError(
            'The required element "{}" was not found in the API response'.format(child_name), node
        )
    return child


def find_text_or_raise(node, child_name):
    """
    Extract the text from the direct descendant of `node` with the tag `child_name`. If the no child is found with that
    tag, an error is raised.

    Args:
        node: an ElementTree.Element object representing a node in an XML document that is expected to have a child
            with the name specified by the `child_name` parameter.
        child_name: a string containing the name of the tag that is expected to be a direct descendant of `node`.

    Raises:
        PaypalMalformedResponseError: If the document does not have the expected structure.

    Returns:
        The text contents of the child when it is found.
    """
    text = node.findtext(child_name)
    if text is None:
        raise PaypalMalformedResponseError(
            'The required element "{}" was not found in the API response'.format(child_name), node
        )
    return text


class PaypalError(Exception):
    """A base class for Paypal API related exceptions."""
    pass


class PaypalApiRequestFailedError(PaypalError):
    """Paypal responded with an error code specifying a generic failure with processing the request."""

    def __init__(self, response_code, response_message, request_type="API"):
        super(PaypalApiRequestFailedError, self).__init__(
            'Paypal {request_type} request failed with code {code}: {message}'.format(
                request_type=request_type,
                code=response_code,
                message=response_message
            )
        )


class PaypalMalformedResponseError(PaypalError):
    """The response from paypal did not match the expected format."""

    def __init__(self, message, root_node=None):
        with_tree = message
        if root_node:
            with_tree = u'{}:{}'.format(
                message,
                ET.tostring(root_node, encoding='UTF-8', method='xml').decode('utf-8'),
            )

        super(PaypalMalformedResponseError, self).__init__(with_tree)


class PaypalApiRequest(object):
    """
    A generic API request. Subclasses are expected to override `append_request_node` to specify the details of the
    particular request in question.

    A particular request expects a particular type of response. The class used to parse the response is specified by
    overriding the `RESPONSE_CLASS` variable.
    """

    # WARNING: the XML parser used by the paypal service will not recognize fields in a different order than the one
    # specified below. DO NOT CHANGE the ordering of fields within requests! They could have avoided this by using
    # <xsd:all> instead of <xsd:sequence> for their types that shouldn't care about the ordering of fields (many of
    # their complex types fall into this category).

    RESPONSE_CLASS = PaypalApiResponse

    def __init__(self, partner, vendor, password, user, url):
        self.partner = partner
        self.vendor = vendor
        self.password = password
        self.user = user
        self.url = url

    def create_request_document(self):
        """Get the string representation of the XML request."""
        root_node = ET.Element('reportingEngineRequest')
        self.append_authentication_node(root_node)
        self.append_request_node(root_node)

        # NOTE: we have to use this API to get the XML declaration, it is suboptimal that we have to construct a
        # BytesIO buffer to write to.
        tree = ET.ElementTree(root_node)
        bytes_buffer = BytesIO()
        tree.write(bytes_buffer, encoding='UTF-8', xml_declaration=True)
        return bytes_buffer.getvalue()

    def append_authentication_node(self, root_node):
        """Inject the authentication elements into the request."""
        auth_node = ET.SubElement(root_node, 'authRequest')

        for attribute in ('user', 'vendor', 'partner', 'password'):
            child_node = ET.SubElement(auth_node, attribute)
            child_node.text = six.text_type(getattr(self, attribute))

    def append_request_node(self, root_node):
        """Inject the request-specific elements into the request."""
        pass

    def execute(self):
        """
        Execute the request and return the parsed response object.

        Returns:
            A subclass of PaypalApiResponse
        """
        request_document = self.create_request_document()
        headers = {
            'Content-Type': 'text/plain'
        }
        response = requests.post(
            self.url,
            data=request_document,
            headers=headers,
        )
        return self.RESPONSE_CLASS.from_http_response(response)


class PaypalReportResponse(PaypalApiResponse):
    """
    A response that details the status of a reporting request. This is used to parse responses from both the
    PaypalReportRequest and PaypalReportResultsRequest request types since they are very similar and they are
    interpreted the same way.

    Args:
        response_code: The numeric status code for the request, this indicates the overall request status.
        response_message: A human readable description of the status of the request.
        report_id: A unique report identifier that can be used to identify the report when making subsequent requests.
        status_code: The numeric status code for the report generation request. Note that this code differs from the
            response_code and refers to the report status, not the request status.
    """

    REPORT_CREATED_STATUS_CODE = 1
    REPORT_EXECUTING_STATUS_CODE = 2
    REPORT_COMPLETE_STATUS_CODE = 3

    def __init__(self, response_code, response_message, report_id, status_code, status_message):
        super(PaypalReportResponse, self).__init__(response_code, response_message)
        self.report_id = report_id
        self.status_code = status_code
        self.status_message = status_message

    @classmethod
    def params_from_xml(cls, root_node):
        maybe_report_node = root_node.find('runReportResponse')
        if maybe_report_node is None:
            maybe_report_node = root_node.find('getResultsResponse')
            if maybe_report_node is not None:
                maybe_report_node = maybe_report_node.find('Results')

        if maybe_report_node is None:
            raise PaypalMalformedResponseError(
                'Unable to find a valid response body, must be either "runReportResponse" or "getResultsResponse".',
                root_node
            )

        params = {}
        params['report_id'] = find_text_or_raise(maybe_report_node, 'reportId')
        params['status_code'] = int(find_text_or_raise(maybe_report_node, 'statusCode'))
        params['status_message'] = find_text_or_raise(maybe_report_node, 'statusMsg')
        return params

    @property
    def is_running(self):
        """Returns True iff the report is still running on the backend."""
        return self.status_code in (self.REPORT_CREATED_STATUS_CODE, self.REPORT_EXECUTING_STATUS_CODE)

    @property
    def is_ready(self):
        """Returns True iff the report completed successfully and data can be retrieved."""
        return self.status_code == self.REPORT_COMPLETE_STATUS_CODE

    def raise_for_status(self):
        """Raise an error if the report failed to execute on the backend."""
        super(PaypalReportResponse, self).raise_for_status()
        if self.status_code > self.REPORT_COMPLETE_STATUS_CODE:
            raise PaypalApiRequestFailedError(self.status_code, self.status_message, "report")


class PaypalReportRequest(PaypalApiRequest):
    """
    Request a report. The different types of reports require different parameters, which are passed into the constructor
    as keyword arguments.

    Args:
        report_name: The name of the report to generate.
        report_params: The parameters to pass along to the backend. This is commonly a date range, but may include other
            types of parameters as well.
    """

    DEFAULT_PAGE_SIZE = 50
    RESPONSE_CLASS = PaypalReportResponse

    def __init__(self, report_name, partner, vendor, password, user, url, **report_params):
        super(PaypalReportRequest, self).__init__(partner, vendor, password, user, url)
        self.report_name = report_name
        self.report_params = report_params
        self.page_size = self.report_params.pop('page_size', self.DEFAULT_PAGE_SIZE)

    def append_request_node(self, root_node):

        # WARNING: the paypal XML parser is position sensitive. Do NOT change the ordering of the fields in the request.
        request_node = ET.SubElement(root_node, 'runReportRequest')
        name_node = ET.SubElement(request_node, 'reportName')
        name_node.text = six.text_type(self.report_name)

        for param_name, param_value in six.iteritems(self.report_params):
            param_node = ET.SubElement(request_node, 'reportParam')
            param_name_node = ET.SubElement(param_node, 'paramName')
            param_name_node.text = six.text_type(param_name)
            param_value_node = ET.SubElement(param_node, 'paramValue')
            param_value_node.text = six.text_type(param_value)

        page_size_node = ET.SubElement(request_node, 'pageSize')
        page_size_node.text = six.text_type(self.page_size)


ColumnMetadata = namedtuple('ColumnMetadata', ('name', 'data_type'))  # pylint: disable=invalid-name


class PaypalReportMetadataResponse(PaypalApiResponse):
    """
    A response that details the format and size of the generated report.

    Args:
        response_code: The numeric status code for the request, this indicates the overall request status.
        response_message: A human readable description of the status of the request.
        num_rows: The number of rows in the report.
        num_pages: The number of pages of data that can be requested.
        page_size: The number of rows per page of data.
        columns: An OrderedDict mapping column names to ColumnMetadata objects.
    """

    def __init__(self, response_code, response_message, num_rows, num_pages, page_size, columns):
        super(PaypalReportMetadataResponse, self).__init__(response_code, response_message)
        self.num_rows = num_rows
        self.num_pages = num_pages
        self.page_size = page_size
        self.columns = columns

    @classmethod
    def params_from_xml(cls, root_node):
        node = find_or_raise(root_node, 'getMetaDataResponse')

        params = {}
        params['num_rows'] = int(find_text_or_raise(node, 'numberOfRows'))
        params['num_pages'] = int(find_text_or_raise(node, 'numberOfPages'))
        params['page_size'] = int(find_text_or_raise(node, 'pageSize'))

        columns = OrderedDict()
        for column_node in node.iterfind('columnMetaData'):
            name = find_text_or_raise(column_node, 'dataName')
            data_type = find_text_or_raise(column_node, 'dataType')
            columns[name] = ColumnMetadata(name=name, data_type=data_type)
        params['columns'] = columns

        return params


class PaypalReportMetadataRequest(PaypalApiRequest):
    """
    A request for the format and size of a particular report.

    Args:
        report_id: The report identifier.
    """

    RESPONSE_CLASS = PaypalReportMetadataResponse

    def __init__(self, report_id, partner, vendor, password, user, url):
        super(PaypalReportMetadataRequest, self).__init__(partner, vendor, password, user, url)
        self.report_id = report_id

    def append_request_node(self, root_node):
        request_node = ET.SubElement(root_node, 'getMetaDataRequest')
        report_id_node = ET.SubElement(request_node, 'reportId')
        report_id_node.text = six.text_type(self.report_id)


class PaypalReportDataResponse(PaypalApiResponse):
    """
    A page of data.

    Args:
        response_code: The numeric status code for the request, this indicates the overall request status.
        response_message: A human readable description of the status of the request.
        rows: A 2-dimensional array that represents the data table.
    """

    def __init__(self, response_code, response_message, rows):
        super(PaypalReportDataResponse, self).__init__(response_code, response_message)
        self.rows = rows

    @classmethod
    def params_from_xml(cls, root_node):
        node = find_or_raise(root_node, 'getDataResponse')

        params = {}
        rows = []
        for row_node in node.iterfind('reportDataRow'):
            row = []
            for column_node in row_node.iterfind('columnData'):
                # NOTE: data types are not used to parse the text data here!
                row.append(find_text_or_raise(column_node, 'data'))

            rows.append(row)

        params['rows'] = rows
        return params


class PaypalReportDataRequest(PaypalApiRequest):
    """
    Request a page of data.

    Args:
        report_id: The report identifier.
        page_num: The page of data to get.
    """

    RESPONSE_CLASS = PaypalReportDataResponse

    def __init__(self, report_id, page_num, partner, vendor, password, user, url):
        super(PaypalReportDataRequest, self).__init__(partner, vendor, password, user, url)
        self.report_id = report_id
        self.page_num = page_num

    def append_request_node(self, root_node):
        request_node = ET.SubElement(root_node, 'getDataRequest')
        report_id_node = ET.SubElement(request_node, 'reportId')
        report_id_node.text = six.text_type(self.report_id)
        page_num_node = ET.SubElement(request_node, 'pageNum')
        page_num_node.text = six.text_type(self.page_num)


class PaypalReportResultsRequest(PaypalApiRequest):
    """
    Request the status of a report.

    Args:
        report_id: The report identifier.
    """

    RESPONSE_CLASS = PaypalReportResponse

    def __init__(self, report_id, partner, vendor, password, user, url):
        super(PaypalReportResultsRequest, self).__init__(partner, vendor, password, user, url)
        self.report_id = report_id

    def append_request_node(self, root_node):
        request_node = ET.SubElement(root_node, 'getResultsRequest')
        report_id_node = ET.SubElement(request_node, 'reportId')
        report_id_node.text = six.text_type(self.report_id)


BaseSettlementReportRecord = namedtuple('SettlementReportRecord', [  # pylint: disable=invalid-name
    'transaction_id',
    'time',
    'type',
    'tender_type',
    'account_number',
    'expires',
    'amount',
    'result_code',
    'response_msg',
    'comment_1',
    'comment_2',
    'batch_id',
    'currency_symbol',
    'paypal_transaction_id',
    'paypal_fees',
    'paypal_email_id',
    'original_pnref',
    'original_type',
    'original_amount',
    'original_time',
    'invoice_number',
    'purchase_order',
    'customer_ref'
])


class SettlementReportRecord(BaseSettlementReportRecord):
    """A record in the settlement report."""
    # pylint: disable=no-member

    @property
    def transaction_type(self):
        """Internal transaction type code from the paypal transaction type code"""
        if self.type == 'Sale':
            return 'sale'
        elif self.type == 'Credit':
            return 'refund'
        else:
            raise TypeError("Unknown transaction type: {0}".format(self.type))

    @property
    def decimal_amount(self):
        """A string representing the amount of currency as a floating point number"""
        return str(self.amount_string_to_decimal(self.amount))

    @property
    def decimal_fees(self):
        """A string representing the amount of currency of the fee as a floating point number"""
        return str(self.amount_string_to_decimal(self.paypal_fees))

    def amount_string_to_decimal(self, amount):
        """
        Paypal represents currency as whole numbers. $50.00 will be represented as "5000" in the report since different
        locales will use different decimal place separators etc. We do two things here:

        1) Ensure the decimal value uses 2 fractional decimal places by appending the string ".00" to the input string
        2) Divide by 100 to get the dollar value amount

        The sequence looks like: "5000" -> "5000.00" -> "50.00"

        By convention, refunds have a negative dollar value.
        """
        decimal_amount = Decimal(amount + '.00') / Decimal(100)
        if self.transaction_type == 'refund':
            decimal_amount = decimal_amount * Decimal(-1)

        return decimal_amount


class PaypalTimeoutError(PaypalError):
    """The requested report did not finish generating in time."""

    def __init__(self, start_time):
        super(PaypalTimeoutError, self).__init__(
            "Aborting since the report did not finish generating within the acceptable time range. Started generation"
            " at {start_time}.".format(
                start_time=start_time
            )
        )



import mock
import pytest
from prefect.core import Flow
from prefect.engine import signals
from pytest_mock import mocker  # noqa: F401

from edx_prefectutils import mysql as utils_mysql


@pytest.fixture
def mock_mysql_connection(mocker):  # noqa: F811
    # Mock the Snowflake connection and cursor.
    mocker.patch.object(utils_mysql, 'create_mysql_connection')
    mock_cursor = mocker.Mock()
    mock_connection = mocker.Mock()
    mock_connection.cursor.return_value = mock_cursor
    utils_mysql.create_mysql_connection.return_value = mock_connection
    return mock_connection


def test_load_s3_data_to_mysql_no_overwrite_existing_data(mock_mysql_connection):
    mock_cursor = mock_mysql_connection.cursor()
    mock_fetchone = mock.Mock()
    mock_cursor.fetchone = mock_fetchone

    task = utils_mysql.load_s3_data_to_mysql
    with pytest.raises(
        signals.SKIP,
        match="Skipping task as data already exists in the dest. table and no overwrite was provided."
    ):
        task.run(
            aurora_credentials={},
            database="test_database",
            table="test_table",
            table_columns=[('id', 'int'), ('course_id', 'varchar(255) NOT NULL')],
            s3_url="s3://edx-test/test/",
            overwrite=False
        )


def test_load_s3_data_to_mysql_overwrite_without_record_filter(mock_mysql_connection):
    mock_cursor = mock_mysql_connection.cursor()
    mock_fetchone = mock.Mock()
    mock_cursor.fetchone = mock_fetchone

    with Flow("test") as f:
        utils_mysql.load_s3_data_to_mysql(
            aurora_credentials={},
            database="test_database",
            table="test_table",
            table_columns=[('id', 'int'), ('course_id', 'varchar(255) NOT NULL')],
            s3_url="s3://edx-test/test/",
            overwrite=True
        )

    state = f.run()
    assert state.is_successful()
    mock_cursor.execute.assert_has_calls(
        [
            mock.call("SELECT 1 FROM test_table  LIMIT 1"),
            mock.call("DELETE FROM test_table ")
        ]
    )


def test_load_s3_data_to_mysql_overwrite_with_record_filter(mock_mysql_connection):
    mock_cursor = mock_mysql_connection.cursor()
    mock_fetchone = mock.Mock()
    mock_cursor.fetchone = mock_fetchone

    with Flow("test") as f:
        utils_mysql.load_s3_data_to_mysql(
            aurora_credentials={},
            database="test_database",
            table="test_table",
            table_columns=[('id', 'int'), ('course_id', 'varchar(255) NOT NULL')],
            s3_url="s3://edx-test/test/",
            record_filter="where course_id='edX/Open_DemoX/edx_demo_course'",
            overwrite=True
        )

    state = f.run()
    assert state.is_successful()
    mock_cursor.execute.assert_has_calls(
        [
            mock.call("SELECT 1 FROM test_table where course_id='edX/Open_DemoX/edx_demo_course' LIMIT 1"),
            mock.call("DELETE FROM test_table where course_id='edX/Open_DemoX/edx_demo_course'")
        ]
    )


def test_load_s3_data_to_mysql(mock_mysql_connection):
    mock_cursor = mock_mysql_connection.cursor()
    mock_fetchone = mock.Mock()
    mock_cursor.fetchone = mock_fetchone

    with Flow("test") as f:
        utils_mysql.load_s3_data_to_mysql(
            aurora_credentials={},
            database="test_database",
            table="test_table",
            table_columns=[('id', 'int'), ('course_id', 'varchar(255) NOT NULL')],
            s3_url="s3://edx-test/test/",
            record_filter="where course_id='edX/Open_DemoX/edx_demo_course'",
            ignore_num_lines=2,
            overwrite=True
        )

    state = f.run()
    assert state.is_successful()
    mock_cursor.execute.assert_has_calls(
        [
            mock.call("\n        CREATE TABLE IF NOT EXISTS test_table (id int,course_id varchar(255) NOT NULL)\n    "), # noqa
            mock.call("SELECT 1 FROM test_table where course_id='edX/Open_DemoX/edx_demo_course' LIMIT 1"), # noqa
            mock.call("DELETE FROM test_table where course_id='edX/Open_DemoX/edx_demo_course'"), # noqa
            mock.call("\n            LOAD DATA FROM S3 PREFIX 's3://edx-test/test/'\n            INTO TABLE test_table\n            FIELDS TERMINATED BY ',' OPTIONALLY ENCLOSED BY ''\n            ESCAPED BY '\\\\'\n            IGNORE 2 LINES\n        "), # noqa
        ]
    )


def test_load_s3_data_to_mysql_overwrite_with_temp_table(mock_mysql_connection):
    mock_cursor = mock_mysql_connection.cursor()

    with Flow("test") as f:
        utils_mysql.load_s3_data_to_mysql(
            aurora_credentials={},
            database="test_database",
            table="test_table",
            table_columns=[('id', 'int'), ('course_id', 'varchar(255) NOT NULL')],
            s3_url="s3://edx-test/test/",
            overwrite=True,
            overwrite_with_temp_table=True,
        )

    state = f.run()
    assert state.is_successful()
    mock_cursor.execute.assert_has_calls(
        [
            mock.call("\n        CREATE TABLE IF NOT EXISTS test_table (id int,course_id varchar(255) NOT NULL)\n    "),
            mock.call("SELECT 1 FROM test_table  LIMIT 1"),
            mock.call("DROP TABLE IF EXISTS test_table_old"),
            mock.call("DROP TABLE IF EXISTS test_table_temp"),
            mock.call("CREATE TABLE test_table_temp (id int,course_id varchar(255) NOT NULL)"),
            mock.call("\n            LOAD DATA FROM S3 PREFIX 's3://edx-test/test/'\n            INTO TABLE test_table_temp\n            FIELDS TERMINATED BY ',' OPTIONALLY ENCLOSED BY ''\n            ESCAPED BY '\\\\'\n            IGNORE 0 LINES\n        "), # noqa
            mock.call("RENAME TABLE test_table to test_table_old, test_table_temp to test_table"),
            mock.call("DROP TABLE IF EXISTS test_table_old"),
            mock.call("DROP TABLE IF EXISTS test_table_temp"),
        ]
    )


def test_table_creation_with_indexes(mock_mysql_connection):
    mock_cursor = mock_mysql_connection.cursor()
    with Flow("test") as f:
        utils_mysql.load_s3_data_to_mysql(
            aurora_credentials={},
            database="test_database",
            table="test_table",
            table_columns=[('user_id', 'int'), ('course_id', 'varchar(255) NOT NULL')],
            table_indexes=[('user_id', ), ('course_id', ), ('user_id', 'course_id')],
            s3_url="s3://edx-test/test/",
            overwrite=True,
            overwrite_with_temp_table=True,
        )

    state = f.run()
    assert state.is_successful()
    mock_cursor.execute.assert_has_calls(
        [
            mock.call("\n        CREATE TABLE IF NOT EXISTS test_table (user_id int,course_id varchar(255) NOT NULL,INDEX (user_id),INDEX (course_id),INDEX (user_id,course_id))\n    "), # noqa
        ]
    )


def test_load_s3_data_to_mysql_with_manifest(mock_mysql_connection):
    mock_cursor = mock_mysql_connection.cursor()
    with Flow("test") as f:
        utils_mysql.load_s3_data_to_mysql(
            aurora_credentials={},
            database="test_database",
            table="test_table",
            table_columns=[('id', 'int'), ('course_id', 'varchar(255) NOT NULL')],
            s3_url="s3://edx-test/some/prefix/",
            overwrite=True,
            overwrite_with_temp_table=True,
            use_manifest=True,
        )

    state = f.run()
    assert state.is_successful()
    mock_cursor.execute.assert_has_calls(
        [
            mock.call("\n            LOAD DATA FROM S3 MANIFEST 's3://edx-test/some/prefix/manifest.json'\n            INTO TABLE test_table_temp\n            FIELDS TERMINATED BY ',' OPTIONALLY ENCLOSED BY ''\n            ESCAPED BY '\\\\'\n            IGNORE 0 LINES\n        "), # noqa
        ]
    )



"""
S3 related common methods and tasks for Prefect
"""

import prefect
from prefect import task
from prefect.tasks.aws import s3
from prefect.utilities.aws import get_boto_client


@task
def delete_s3_directory(bucket: str = None, prefix: str = None, credentials: dict = None):
    """
    Deletes all objects with the given S3 directory (prefix) from the given bucket.

    Args:
        bucket (str): The S3 bucket to delete the objects from.
        prefix (str): The S3 prefix to delete the objects from.
        credentials (dict): The AWS credentials to use.
    """
    s3_keys = list_object_keys_from_s3.run(bucket, prefix, credentials)
    if s3_keys:
        s3_client = get_boto_client('s3', credentials=credentials)
        logger = prefect.context.get("logger")
        logger.info("Deleting S3 keys: {}".format(s3_keys))
        s3_client.delete_objects(
            Bucket=bucket,
            Delete={
                'Objects': [{'Key': key} for key in s3_keys]
            }
        )


@task
def delete_object_from_s3(key: str = None, bucket: str = None, credentials: dict = None, ):
    """
    Delete an object from S3.

    key (str): Name of the object within the S3 bucket (/foo/bar/baz.json)
    bucket (str): Name of the S3 bucket to delete from.
    credentials (dict): AWS credentials, if None boto will fall back the usual methods of resolution.
    """
    s3_client = get_boto_client("s3", credentials=credentials)
    s3_client.delete_object(Bucket=bucket, Key=key)


@task
def list_object_keys_from_s3(bucket: str = None, prefix: str = '', credentials: dict = None, ):
    """
    List objects key names from an S3 bucket that match the given prefix.

    prefix (str): Prefix path to search (ex: /foo/bar will match /foo/bar/baz and /foo/bar/baz/bing ...)
    bucket (str): Name of the S3 bucket to search from.
    credentials (dict): AWS credentials, if None boto will fall back the usual methods of resolution.
    """
    s3_client = get_boto_client("s3", credentials=credentials)
    found_objects = s3_client.list_objects_v2(Bucket=bucket, Prefix=prefix)

    logger = prefect.context.get("logger")
    logger.info(found_objects)

    if found_objects['KeyCount']:
        return [
            o['Key'] for o in found_objects['Contents']
        ]
    else:
        return []


def get_s3_path_for_date(date):
    # The path and file name inside our given bucket and S3 prefix to write the file to
    return '{date}/{date}.json'.format(date=date)


@task
def write_report_to_s3(download_results: tuple, s3_bucket: str, s3_path: str):
    logger = prefect.context.get("logger")

    date, report_str = download_results
    date_path = get_s3_path_for_date(date)
    s3_key = s3_path + date_path
    logger.info("Writing report to S3 for {} to {}".format(date, s3_key))

    s3.S3Upload(bucket=s3_bucket).run(
        report_str,
        key=s3_key
    )

    return date_path


@task
def get_s3_url(s3_bucket, s3_path):
    return 's3://{bucket}/{path}'.format(bucket=s3_bucket, path=s3_path)



"""A simple client for authenticated access to Open edX REST APIs."""

from datetime import datetime, timedelta

import backoff
import requests
from prefect.utilities.logging import get_logger
from requests.auth import AuthBase

DEFAULT_RETRY_STATUS_CODES = (
    requests.codes.request_timeout,         # HTTP Status Code 408
    requests.codes.too_many_requests,       # HTTP Status Code 429
    requests.codes.service_unavailable,     # HTTP Status Code 503
    requests.codes.gateway_timeout,         # HTTP Status Code 504
    requests.codes.bad_gateway,             # HTTP Status Code 502
    520,                                    # This is a custom Cloudwatch code for "Unknown error".
)
DEFAULT_TIMEOUT_SECONDS = 7200


class EdxApiClient(object):
    """
    Simplifies authentication and pagination logic when communicating with Open edX REST APIs.

    Arguments:
        auth_url (str): The full URL of the authentication endpoint. It should look like
            https://example.com/oauth2/v1/access_token. If this is not provided it is read from the "auth_url" field
            of the "edx-rest-api" section of the configuration file.
        client_id (str): This is the OAuth2 client_id for the application. If it is not provided it is read from the
            "client_id" field of the "edx-rest-api" section of the configuration file.
        client_secret (str): This is the OAuth2 client_secret for the application. If it is not provided it is read from
            the "client_secret" field of the "edx-rest-api" section of the configuration file.
        token_type (str): The type of authentication token required for the API call.  Should be one of 'jwt' (default)
            or 'bearer'.
    """

    def __init__(self, auth_url=None,
                 client_id=None, client_secret=None,
                 token_type=None):

        self._expires_at = None
        self._session = requests.Session()
        self._session.hooks = {
            'response': log_response_hook
        }

        self.client_id = client_id
        self.client_secret = client_secret
        self.auth_url = auth_url
        self.token_type = token_type or 'jwt'

    @property
    def authenticated_session(self):
        """A session that has a valid access token associated with it and can make authenticated requests."""
        self.ensure_oauth_access_token()
        return self._session

    def ensure_oauth_access_token(self):
        """Retrieves OAuth 2.0 access token using the client credentials grant and stores it in the request session."""
        logger = get_logger()
        now = datetime.utcnow()
        if self._expires_at is None or now >= self._expires_at:
            logger.info('Token is expired or missing, requesting a new one.')

            data = {
                'grant_type': 'client_credentials',
                'client_id': self.client_id,
                'client_secret': self.client_secret,
                'token_type': self.token_type,
            }

            response = requests.post(
                self.auth_url,
                data=data,
                hooks={
                    'response': log_response_hook
                }
            )
            data = response.json()
            self._session.auth = SuppliedAuth(data['access_token'], data.get('token_type', self.token_type))
            self._expires_at = now + timedelta(seconds=data['expires_in'])
            logger.info("Acquired a token that expires at {}".format(self._expires_at.isoformat()))

    def get(self, url, params=None, timeout_seconds=DEFAULT_TIMEOUT_SECONDS, retry_on=DEFAULT_RETRY_STATUS_CODES):
        """
        Fetches a single page of the given resource.

        Arguments:
            url (str): The URL of the resource.
            params (dict): This is a dictionary of key-value pairs that are URL encoded and injected into the query
                string when making the request.
            timeout_seconds (float): When requesting a page, keep retrying unless this much time has elapsed. This timer
                resets after every successful fetch of a page. Raise an error if it takes longer than this amount of
                time to fetch an individual page.
            retry_on (iterable): This is a set of HTTP status codes that should trigger a retry of the request if they
                are received from the server in the response. If one is received the system implements an exponential
                back-off and repeatedly requests the page until either the timeout expires, a fatal exception occurs, or
                an OK response is received.

        Returns: A single requests.Response object for the first page of data received from the server.
        """
        return next(self.paginated_get(url, params=params, timeout_seconds=timeout_seconds, retry_on=retry_on,
                                       pagination_key=None))

    def paginated_get(self, url, params=None, timeout_seconds=DEFAULT_TIMEOUT_SECONDS,
                      retry_on=DEFAULT_RETRY_STATUS_CODES, pagination_key='next'):
        """
        Fetches a paginated resource.

        Arguments:
            url (str): The URL of the resource.
            params (dict): This is a dictionary of key-value pairs that are URL encoded and injected into the query
                string when making the request.
            timeout_seconds (float): When requesting a page, keep retrying unless this much time has elapsed. This timer
                resets after every successful fetch of a page. Raise an error if it takes longer than this amount of
                time to fetch an individual page.
            retry_on (iterable): This is a set of HTTP status codes that should trigger a retry of the request if they
                are received from the server in the response. If one is received the system implements an exponential
                back-off and repeatedly requests the page until either the timeout expires, a fatal exception occurs, or
                an OK response is received.
            pagination_key (str or lambda): Used to locate the URL for the next page of data from the JSON-parsed
                response.  If "pagination_key" is a string, then it names a field at the root of the response containing
                the "next page" URL.  E.g., use "pagination_key='next'" for a response that looks like this:
                {"results": [...], "next": "http://...", "previous": "http://..."}

                If "pagination_key" is a lambda, it should return the next page url from the given response object.
                E.g., use "pagination_key=lambda r: r['pagination']['next']" for a response that looks like this:

                {"results": [...], "pagination": {"next": "http://...", "previous": "http://..."}}

        Yields: A single requests.Response object for each page of data received from the server.
        """

        def get_next_url_from_response(response):
            """Returns the next page's URL from the response, as located by pagination_key."""
            response_obj = response.json()
            if isinstance(pagination_key, str):
                return response_obj.get(pagination_key)
            elif callable(pagination_key):
                return pagination_key(response_obj)
            else:
                return None

        def should_giveup(error):
            """
            Give up if the response is `None` or if the status code is not in the set of status codes
            that are retryable.
            """
            error_response = getattr(error, 'response', None)
            if error_response is None:
                return True

            return error_response.status_code not in retry_on

        @backoff.on_exception(backoff.expo,
                              requests.exceptions.RequestException,
                              max_time=timeout_seconds,
                              giveup=should_giveup)
        def get_resource_with_retry(next_url=None):
            """
            Attempt to get the resource, using an exponetial back-off to retry recoverable, failed requests.

            Arguments:
                next_url (str): The url of the next page to fetch. If this is `None` this is the first page, so we
                    should use the provided `url` and `params` passed into the outer function to fetch this first page.
                    The next url is provided in the first response with all of the appropriate parameters needed to
                    fetch the next page of data. We don't want to accidentally override existing parameters so we omit
                    the `params` kwarg from the call.
            """
            if next_url is None:
                raw_response = self.authenticated_session.get(url, params=params)
            else:
                raw_response = self.authenticated_session.get(next_url)

            raw_response.raise_for_status()

            # Get next URL if pagination was requested
            next_url = get_next_url_from_response(raw_response)

            return raw_response, next_url

        next_url = None
        while True:
            response, next_url = get_resource_with_retry(next_url)
            yield response
            if next_url is None:
                break


class SuppliedAuth(AuthBase):
    """Attaches a supplied authentication to the given Request object."""

    def __init__(self, token, token_type):
        """Instantiate the auth class."""
        self.token = token
        self.token_type = token_type

    def __call__(self, r):
        """Update the request headers."""
        r.headers['Authorization'] = '{token_type} {token}'.format(token_type=self.token_type, token=self.token)
        return r


def log_response_hook(response, *args, **kwargs):  # pylint: disable=unused-argument
    """Log summary information about every request made."""
    logger = get_logger()
    logger.info(
        "[{}] [{}] [{}] {}".format(
            response.request.method, response.status_code, response.elapsed.total_seconds(), response.url
        )
    )



#!/usr/bin/env python

"""
Tests for Hashicorp Vault secrets utils in the `edx_prefectutils` package.
"""

from prefect import Flow, task, unmapped
from pytest_mock import mocker  # noqa: F401

from edx_prefectutils import vault_secrets


@task
def get_val(secret):
    return secret.get("test")


def test_read_vault_secret(mocker):  # noqa: F811
    mocker.patch.object(vault_secrets, 'open')
    mocker.patch.object(vault_secrets.hvac, 'Client')
    with Flow("test") as f:
        secret_val = vault_secrets.VaultKVSecret(
            path="warehouses/test_platform/test_secret",
            version=2
        )
        get_val(unmapped(secret_val))
    state = f.run()
    assert state.is_successful()



import xml.etree.cElementTree as ET
from collections import OrderedDict
from io import StringIO
from unittest import TestCase

import httpretty
from ddt import data, ddt, unpack

from edx_prefectutils.paypal_xml import (ColumnMetadata,
                                         PaypalApiRequestFailedError,
                                         PaypalMalformedResponseError,
                                         PaypalReportDataRequest,
                                         PaypalReportMetadataRequest,
                                         PaypalReportRequest,
                                         PaypalReportResultsRequest,
                                         SettlementReportRecord)

TEST_URL = 'http://test.api/endpoint'


class XmlRequestMixin(object):

    SAMPLE_RESPONSE = ''

    def setUp(self):
        super(XmlRequestMixin, self).setUp()
        self.response_xml_root = ET.fromstring(self.SAMPLE_RESPONSE)

    def on_post_return_xml(self):
        element_tree = ET.ElementTree(self.response_xml_root)
        string_buffer = StringIO()
        element_tree.write(string_buffer, encoding='unicode', xml_declaration=True)
        response_xml_root_string = string_buffer.getvalue()
        httpretty.register_uri(httpretty.POST, TEST_URL, response_xml_root_string)

    def remove_xml_node(self, path):
        element = self.response_xml_root.findall(path)[0]
        parent = self.response_xml_root.findall(path + '/..')[0]
        parent.remove(element)

    def set_xml_node_text(self, path, value):
        element = self.response_xml_root.findall(path)[0]
        element.text = str(value)

    def parse_request_xml(self):
        http_request = httpretty.last_request()
        self.assertEqual(http_request.method, "POST")
        self.assertEqual(http_request.headers['Content-type'], 'text/plain')
        self.request_xml_root = ET.fromstring(http_request.body)

    def assert_request_xml_equals(self, path, expected_xml):
        element = self.request_xml_root.findall(path)[0]
        expected_root = ET.fromstring(expected_xml)
        self.assert_nodes_equal(expected_root, element, element.tag)

    def assert_nodes_equal(self, expected, actual, path):
        if len(actual) == 0:
            self.assertEqual(
                expected.text,
                actual.text,
                "Text does not match at path '{path}', expected: {exp}, actual: {act}".format(
                    path=path,
                    exp=expected.text,
                    act=actual.text
                )
            )
        expected_children_tags = [c.tag for c in expected]
        actual_children_tags = [c.tag for c in actual]

        self.assertEqual(
            expected_children_tags,
            actual_children_tags,
            "Children do not match at path '{path}', expected: {exp}, actual: {act}".format(
                path=path,
                exp=expected_children_tags,
                act=actual_children_tags
            )
        )
        for expected_child, actual_child in zip(expected, actual):
            self.assert_nodes_equal(expected_child, actual_child, path + '/' + expected_child.tag)


@ddt
@httpretty.activate
class TestReportRequest(XmlRequestMixin, TestCase):

    SAMPLE_RESPONSE = """\
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<reportingEngineResponse>
    <baseResponse>
        <responseCode>100</responseCode>
        <responseMsg>Request has completed successfully</responseMsg>
    </baseResponse>
    <runReportResponse>
        <reportId>RE12345690123</reportId>
        <statusCode>3</statusCode>
        <statusMsg>Report has completed successfully</statusMsg>
    </runReportResponse>
</reportingEngineResponse>
"""

    def create_report_request(self):
        return PaypalReportRequest(
            "FooReport",
            "testpartner",
            "testvendor",
            "notsosecret",
            "testuser",
            TEST_URL,
            report_date="2015-08-28"
        )

    def test_complete_status(self):
        self.on_post_return_xml()
        report_request = self.create_report_request()
        response = report_request.execute()

        self.assertEqual(response.response_code, 100)
        self.assertEqual(response.response_message, 'Request has completed successfully')
        self.assertEqual(response.report_id, 'RE12345690123')
        self.assertEqual(response.status_code, 3)
        self.assertEqual(response.status_message, 'Report has completed successfully')
        self.assertFalse(response.is_running)
        self.assertTrue(response.is_ready)

        self.parse_request_xml()
        self.assert_request_xml_equals('.', """\
<reportingEngineRequest>
    <authRequest>
        <user>testuser</user>
        <vendor>testvendor</vendor>
        <partner>testpartner</partner>
        <password>notsosecret</password>
    </authRequest>
    <runReportRequest>
        <reportName>FooReport</reportName>
        <reportParam>
            <paramName>report_date</paramName>
            <paramValue>2015-08-28</paramValue>
        </reportParam>
        <pageSize>50</pageSize>
    </runReportRequest>
</reportingEngineRequest>
""")

    @data(
        (1, 'Report has been created'),
        (2, 'Report is currently executing')
    )
    @unpack
    def test_executing_status(self, status_code, status_msg):
        self.set_xml_node_text('./runReportResponse/statusCode', status_code)
        self.set_xml_node_text('./runReportResponse/statusMsg', status_msg)
        self.on_post_return_xml()

        report_request = self.create_report_request()
        response = report_request.execute()

        self.assertEqual(response.status_code, status_code)
        self.assertEqual(response.status_message, status_msg)
        self.assertTrue(response.is_running)
        self.assertFalse(response.is_ready)

    def test_missing_base_response(self):
        self.remove_xml_node('./baseResponse')
        self.on_post_return_xml()

        report_request = self.create_report_request()
        with self.assertRaisesRegex(PaypalMalformedResponseError, 'baseResponse'):
            report_request.execute()

    @data('responseCode', 'responseMsg')
    def test_missing_base_response_field(self, field_name):
        self.remove_xml_node('./baseResponse/' + field_name)
        self.on_post_return_xml()

        report_request = self.create_report_request()
        with self.assertRaisesRegex(PaypalMalformedResponseError, field_name):
            report_request.execute()

    @data(
        (101, 'Request has failed'),
        (102, 'An internal scheduler error has occurred'),
        (103, 'Unknown report requested'),
        (104, 'Invalid Report ID'),
        (105, 'A system error has occurred'),
        (106, 'A database error has occurred'),
        (107, 'Invalid XML request'),
        (108, 'User authentication failed'),
        (109, 'Invalid report parameters provided'),
        (110, 'Invalid merchant account'),
        (111, 'Invalid page number'),
        (112, 'Template already exists'),
        (113, 'Unknown template requested')
    )
    @unpack
    def test_unsuccessful_request(self, response_code, response_msg):
        self.set_xml_node_text('./baseResponse/responseCode', response_code)
        self.set_xml_node_text('./baseResponse/responseMsg', response_msg)
        self.on_post_return_xml()

        report_request = self.create_report_request()
        with self.assertRaisesRegex(
                PaypalApiRequestFailedError, 'API request failed with code {0}: {1}'.format(
                    response_code,
                    response_msg
                )
        ):
            report_request.execute()

    def test_missing_report_response(self):
        self.remove_xml_node('./runReportResponse')
        self.on_post_return_xml()

        report_request = self.create_report_request()
        with self.assertRaisesRegex(PaypalMalformedResponseError, 'runReportResponse'):
            report_request.execute()

    @data(
        (4, 'Report has failed'),
        (5, 'Report has expired'),
        (6, 'Report has expired')
    )
    @unpack
    def test_unsuccessful_report_generation_attempt(self, status_code, status_msg):
        self.set_xml_node_text('./runReportResponse/statusCode', status_code)
        self.set_xml_node_text('./runReportResponse/statusMsg', status_msg)
        self.on_post_return_xml()

        report_request = self.create_report_request()
        with self.assertRaisesRegex(
                PaypalApiRequestFailedError, 'report request failed with code {0}: {1}'.format(
                    status_code,
                    status_msg
                )
        ):
            report_request.execute()

    @data('reportId', 'statusCode', 'statusMsg')
    def test_missing_base_response_field_run_report_Response(self, field_name):
        self.remove_xml_node('./runReportResponse/' + field_name)
        self.on_post_return_xml()

        report_request = self.create_report_request()
        with self.assertRaisesRegex(PaypalMalformedResponseError, field_name):
            report_request.execute()


@ddt
@httpretty.activate
class TestReportMetadataRequest(XmlRequestMixin, TestCase):

    SAMPLE_RESPONSE = """\
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<reportingEngineResponse>
    <baseResponse>
        <responseCode>100</responseCode>
        <responseMsg>Request has completed successfully</responseMsg>
    </baseResponse>
    <getMetaDataResponse>
        <reportId>RE1234567890</reportId>
        <numberOfRows>102</numberOfRows>
        <numberOfPages>3</numberOfPages>
        <pageSize>50</pageSize>
        <numberOfColumns>3</numberOfColumns>
        <columnMetaData colNum="1">
            <dataName>Transaction ID</dataName>
            <dataType>string</dataType>
        </columnMetaData>
        <columnMetaData colNum="2">
            <dataName>Time</dataName>
            <dataType>date</dataType>
        </columnMetaData>
        <columnMetaData colNum="3">
            <dataName>Type</dataName>
            <dataType>string</dataType>
        </columnMetaData>
    </getMetaDataResponse>
</reportingEngineResponse>
"""

    def create_report_metadata_request(self):
        return PaypalReportMetadataRequest(
            "RE1234567890",
            "testpartner",
            "testvendor",
            "notsosecret",
            "testuser",
            TEST_URL
        )

    def test_successful_response(self):
        self.on_post_return_xml()

        request = self.create_report_metadata_request()
        response = request.execute()

        self.assertEqual(response.response_code, 100)
        self.assertEqual(response.response_message, 'Request has completed successfully')
        self.assertEqual(response.num_rows, 102)
        self.assertEqual(response.num_pages, 3)
        self.assertEqual(response.page_size, 50)
        self.assertEqual(
            response.columns.popitem(last=False),
            ('Transaction ID', ColumnMetadata(name='Transaction ID', data_type='string'))
        )
        self.assertEqual(response.columns.popitem(last=False), ('Time', ColumnMetadata(name='Time', data_type='date')))
        self.assertEqual(
            response.columns.popitem(last=False),
            ('Type', ColumnMetadata(name='Type', data_type='string'))
        )

        self.parse_request_xml()
        self.assert_request_xml_equals('./getMetaDataRequest', """\
<getMetaDataRequest>
    <reportId>RE1234567890</reportId>
</getMetaDataRequest>
""")

    @data('numberOfRows', 'numberOfPages', 'pageSize')
    def test_missing_required_node(self, field_name):
        self.remove_xml_node('./getMetaDataResponse/' + field_name)
        self.on_post_return_xml()

        request = self.create_report_metadata_request()
        with self.assertRaisesRegex(PaypalMalformedResponseError, field_name):
            request.execute()

    @data('dataName', 'dataType')
    def test_malformed_column_metadata(self, field_name):
        self.remove_xml_node('./getMetaDataResponse/columnMetaData[@colNum="1"]/' + field_name)
        self.on_post_return_xml()

        request = self.create_report_metadata_request()
        with self.assertRaisesRegex(PaypalMalformedResponseError, field_name):
            request.execute()

    def test_missing_metadata_response(self):
        self.remove_xml_node('./getMetaDataResponse')
        self.on_post_return_xml()

        request = self.create_report_metadata_request()
        with self.assertRaisesRegex(PaypalMalformedResponseError, 'getMetaDataResponse'):
            request.execute()


@ddt
@httpretty.activate
class TestReportDataRequest(XmlRequestMixin, TestCase):

    SAMPLE_RESPONSE = """\
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<reportingEngineResponse>
    <baseResponse>
        <responseCode>100</responseCode>
        <responseMsg>Request has completed successfully</responseMsg>
    </baseResponse>
    <getDataResponse>
        <reportId>RE1234567890</reportId>
        <reportDataRow rowNum="1">
            <columnData colNum="1">
                <data>ABTCRF9KLMZQ</data>
            </columnData>
            <columnData colNum="2">
                <data>2015-08-27 00:14:27</data>
            </columnData>
            <columnData colNum="3">
                <data>Sale</data>
            </columnData>
        </reportDataRow>
        <reportDataRow rowNum="2">
            <columnData colNum="1">
                <data>ABTMMK7332AR</data>
            </columnData>
            <columnData colNum="2">
                <data>2015-08-27 00:59:05</data>
            </columnData>
            <columnData colNum="3">
                <data/>
            </columnData>
        </reportDataRow>
        <pageNum>1</pageNum>
    </getDataResponse>
</reportingEngineResponse>
"""

    def create_report_data_request(self):
        return PaypalReportDataRequest(
            "RE1234567890",
            1,
            "testpartner",
            "testvendor",
            "notsosecret",
            "testuser",
            TEST_URL
        )

    def test_successful_response(self):
        self.on_post_return_xml()

        request = self.create_report_data_request()
        response = request.execute()

        self.assertEqual(response.response_code, 100)
        self.assertEqual(response.response_message, 'Request has completed successfully')
        self.assertEqual(response.rows, [
            ['ABTCRF9KLMZQ', '2015-08-27 00:14:27', 'Sale'],
            ['ABTMMK7332AR', '2015-08-27 00:59:05', ''],
        ])

        self.parse_request_xml()
        self.assert_request_xml_equals('./getDataRequest', """\
<getDataRequest>
    <reportId>RE1234567890</reportId>
    <pageNum>1</pageNum>
</getDataRequest>
""")

    def test_missing_data_element(self):
        self.remove_xml_node('./getDataResponse/reportDataRow[@rowNum="1"]/columnData[@colNum="1"]/data')
        self.on_post_return_xml()

        request = self.create_report_data_request()
        with self.assertRaisesRegex(PaypalMalformedResponseError, 'data'):
            request.execute()

    def test_missing_data_response(self):
        self.remove_xml_node('./getDataResponse')
        self.on_post_return_xml()

        request = self.create_report_data_request()
        with self.assertRaisesRegex(PaypalMalformedResponseError, 'getDataResponse'):
            request.execute()


@ddt
@httpretty.activate
class TestReportResultsRequest(XmlRequestMixin, TestCase):

    SAMPLE_RESPONSE = """\
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<reportingEngineResponse>
    <baseResponse>
        <responseCode>100</responseCode>
        <responseMsg>Request has completed successfully</responseMsg>
    </baseResponse>
    <getResultsResponse>
        <Results>
            <reportId>RE1234567890</reportId>
            <statusCode>3</statusCode>
            <statusMsg>Report has completed successfully</statusMsg>
        </Results>
    </getResultsResponse>
</reportingEngineResponse>
"""

    def create_report_results_request(self):
        return PaypalReportResultsRequest(
            "RE1234567890",
            "testpartner",
            "testvendor",
            "notsosecret",
            "testuser",
            TEST_URL
        )

    def test_successful_response(self):
        self.on_post_return_xml()

        report_request = self.create_report_results_request()
        response = report_request.execute()

        self.assertEqual(response.response_code, 100)
        self.assertEqual(response.response_message, 'Request has completed successfully')
        self.assertEqual(response.report_id, 'RE1234567890')
        self.assertEqual(response.status_code, 3)
        self.assertEqual(response.status_message, 'Report has completed successfully')
        self.assertFalse(response.is_running)
        self.assertTrue(response.is_ready)

        self.parse_request_xml()
        self.assert_request_xml_equals('./getResultsRequest', """\
<getResultsRequest>
    <reportId>RE1234567890</reportId>
</getResultsRequest>
""")

    @data(
        (1, 'Report has been created'),
        (2, 'Report is currently executing')
    )
    @unpack
    def test_executing_status(self, status_code, status_msg):
        self.set_xml_node_text('./getResultsResponse/Results/statusCode', status_code)
        self.set_xml_node_text('./getResultsResponse/Results/statusMsg', status_msg)
        self.on_post_return_xml()

        report_request = self.create_report_results_request()
        response = report_request.execute()

        self.assertTrue(response.is_running)
        self.assertFalse(response.is_ready)


SAMPLE_TRANSACTION = OrderedDict([
    ('transaction_id', 'BTP123456'),
    ('time', '2015-08-28 00:14:27'),
    ('type', 'Sale'),
    ('tender_type', 'PayPal'),
    ('account_number', 'foobarbaz@foo.com'),
    ('expires', ''),
    ('amount', '5000'),
    ('result_code', '0'),
    ('response_msg', 'Approved'),
    ('comment_1', ''),
    ('comment_2', ''),
    ('batch_id', '0'),
    ('currency_symbol', 'USD'),
    ('paypal_transaction_id', '1FW12345678901234'),
    ('paypal_fees', '140'),
    ('paypal_email_id', 'foobarbaz@foo.com'),
    ('original_pnref', ''),
    ('original_type', ''),
    ('original_amount', ''),
    ('original_time', ''),
    ('invoice_number', 'EDX-123456'),
    ('purchase_order', ''),
    ('customer_ref', '')
])


@ddt
class TestSettlementReportRecord(TestCase):

    @data(
        ('Sale', 'sale'),
        ('Credit', 'refund')
    )
    @unpack
    def test_transaction_type(self, trans_type, internal_trans_type):
        transaction = OrderedDict(SAMPLE_TRANSACTION)
        transaction['type'] = trans_type

        self.assertEqual(SettlementReportRecord(**transaction).transaction_type, internal_trans_type)

    @data(
        ('0', '0.00'),
        ('-1', '-0.01'),
        ('10', '0.10'),
        ('100', '1.00'),
        ('1000', '10.00')
    )
    @unpack
    def test_converting_to_decimal(self, paypal_representation, decimal_amount):
        transaction = OrderedDict(SAMPLE_TRANSACTION)
        transaction['amount'] = paypal_representation
        transaction['paypal_fees'] = paypal_representation
        record = SettlementReportRecord(**transaction)

        self.assertEqual(record.decimal_amount, decimal_amount)
        self.assertEqual(record.decimal_fees, decimal_amount)

    def test_refund(self):
        transaction = OrderedDict(SAMPLE_TRANSACTION)
        transaction['type'] = 'Credit'
        record = SettlementReportRecord(**transaction)

        self.assertEqual(record.decimal_amount, '-50.00')
        self.assertEqual(record.decimal_fees, '-1.40')




import backoff
import prefect
import requests
from prefect import task
from prefect.backend import get_key_value, set_key_value
from snowflake.connector import ProgrammingError

from . import snowflake


@task
@backoff.on_exception(backoff.expo,
                      ProgrammingError,
                      max_tries=3)
def sync_hubspot_to_braze(
    sf_credentials: snowflake.SFCredentials,
    sf_role: str,
    sf_database: str,
    sf_schema: str,
    braze_api_key: str,
    braze_api_server: str,
):
    logger = prefect.context.get("logger")

    # Also, to reduce churn and data usage, only sync items that have changed since our last successful run
    LASTMODIFIEDDATE_KEY = 'email-unsubscribes-hubspot-lastmodifieddate'
    where = ''
    try:
        lastmodifieddate = get_key_value(key=LASTMODIFIEDDATE_KEY)
        where = f" WHERE lastmodifieddate > '{lastmodifieddate}'"
        logger.info('Read lastmodifieddate of %s', lastmodifieddate)
    except ValueError:
        pass

    sf_connection = snowflake.create_snowflake_connection(
        sf_credentials,
        sf_role,
    )
    sf_table = 'email_unsubscribes_hubspot_to_braze'
    query = """
    SELECT email, lastmodifieddate FROM {table} {where}
    """.format(
        table=snowflake.qualified_table_name(sf_database, sf_schema, sf_table),
        where=where,
    )
    cursor = sf_connection.cursor()
    cursor.execute(query)
    count = 0
    max_date = None
    # 50 email batches because that's how many braze accepts at a time
    batch = cursor.fetchmany(50)
    while len(batch) > 0:
        unsubscribe_emails_braze(
            braze_api_server,
            braze_api_key,
            [row[0] for row in batch]
        )
        count += len(batch)
        if count % 500 == 0:
            logger.info("Unsubscribed %d emails from braze", count)

        max_batch_date = max(row[1] for row in batch)
        max_date = max(max_batch_date, max_date) if max_date else max_batch_date

        batch = cursor.fetchmany(50)

    if max_date:
        logger.info('Saving lastmodifieddate of %s', max_date.isoformat())
        set_key_value(key=LASTMODIFIEDDATE_KEY, value=max_date.isoformat())


# Retry rate limit errors for 5 minutes
@backoff.on_predicate(
    backoff.expo,
    lambda resp: resp.status_code == 429,
    max_time=300
)
# Retry on server error 3 times
@backoff.on_predicate(
    backoff.expo,
    lambda resp: resp.status_code >= 500,
    max_tries=3
)
def unsubscribe_emails_braze(braze_api_server, braze_api_key, emails):
    if len(emails) > 50:
        raise ValueError(
            f"At most 50 emails at a time can be unsubscribed "
            f"to Braze, attempted {len(emails)}"
        )

    return requests.post(
        url=f"{braze_api_server}/email/status",
        json={
            'email': emails,
            'subscription_state': 'unsubscribed',
        },
        headers={'Authorization': f'BEARER {braze_api_key}'}
    )
